# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Task to use blhc in debusine."""

from pathlib import Path
from typing import override

from debian.debian_support import Version

import debusine.tasks.tags as ttags
import debusine.worker.tags as wtags
from debusine import utils
from debusine.artifacts import BlhcArtifact
from debusine.artifacts.models import (
    ArtifactCategory,
    DebianPackageBuildLog,
    WorkRequestResults,
    get_architectures,
    get_source_package_name,
    get_source_package_version,
)
from debusine.client.models import RelationType
from debusine.tasks import BaseTaskWithExecutor, RunCommandTask
from debusine.tasks.inputs import EnvironmentInput, SingleInput, Stage
from debusine.tasks.models import BlhcData, BlhcDynamicData


class Blhc(
    RunCommandTask[BlhcData, BlhcDynamicData],
    BaseTaskWithExecutor[BlhcData, BlhcDynamicData],
):
    """Task to use blhc (build-log hardening check) in debusine."""

    TASK_VERSION = 1

    CAPTURE_OUTPUT_FILENAME = "blhc.txt"

    input_artifact = SingleInput(
        field="input.artifact",
        categories=[ArtifactCategory.PACKAGE_BUILD_LOG],
        stage=Stage.PENDING,
    )
    environment = EnvironmentInput(stage=Stage.PENDING)

    @override
    def post_init(self) -> None:
        super().post_init()

        # list of build logs to be checked
        self._blhc_target: Path | None = None

    @override
    def compute_scheduler_tags(
        self,
    ) -> tuple[ttags.TaskProvidedTags, wtags.TaskRequiredTags]:
        provided, required = super().compute_scheduler_tags()

        assert isinstance(self.input_artifact.data, DebianPackageBuildLog)
        provided.add(
            ttags.ProvenanceProvided.SYSTEM,
            {
                ttags.SOURCE_PACKAGE_PREFIX
                + get_source_package_name(self.input_artifact.data)
            },
        )
        return provided, required

    @override
    def get_subject(self) -> str | None:
        assert isinstance(self.input_artifact.data, DebianPackageBuildLog)
        return get_source_package_name(self.input_artifact.data)

    @override
    def compute_dynamic_data(self) -> BlhcDynamicData:
        """
        Resolve artifact lookups for this task.

        :subject: source package name out of the analyzed build log
        :parameter_summary: source package name and version (arch)
        """
        input_artifact = self.input_artifact
        assert isinstance(input_artifact.data, DebianPackageBuildLog)
        package_name = get_source_package_name(input_artifact.data)
        package_version = get_source_package_version(input_artifact.data)
        architectures = " ".join(sorted(get_architectures(input_artifact.data)))

        return BlhcDynamicData(
            environment_id=self.environment.artifact_id,
            input_artifact_id=input_artifact.artifact_id,
            # We don't need the dose-distcheck analysis, and it's
            # potentially complex.
            input_artifact_data=input_artifact.data.model_copy(
                update={"bd_uninstallable": None}
            ),
            subject=self.get_subject(),
            parameter_summary=f"{package_name}_{package_version} "
            f"({architectures})",
        )

    def _blhc_supports_debian_option(self) -> bool:
        """
        Check whether the environment's blhc supports ``--debian``.

        This was added in https://bugs.debian.org/845339, and we want to use
        it if it's available, but if it isn't then it's better to omit that
        option than to fail.
        """
        return self.get_package_version("blhc") >= Version(
            "0.07+20170817+gita232d32"
        )

    @override
    def _cmdline(self) -> list[str]:
        """
        Build the blhc command line.

        Use configuration of self.data and self.blhc_target.
        """
        cmd = ["blhc"]
        if self._blhc_supports_debian_option():
            cmd.append("--debian")

        if extra_flags := self.data.extra_flags:
            # we already checked that the flags are sane...
            cmd += extra_flags

        cmd.append(str(self._blhc_target))

        return cmd

    @override
    def fetch_input(self, destination: Path) -> bool:
        """Download the required artifacts."""
        assert self.dynamic_data

        self.fetch_artifact(self.dynamic_data.input_artifact_id, destination)

        return True

    @override
    def configure_for_execution(self, download_directory: Path) -> bool:
        """
        Find .build files in the input artifacts.

        Set self._blhc_target to the relevant file.

        :param download_directory: where to search the files
        :return: True if valid files were found
        """
        self._prepare_executor_instance()

        if self.executor_instance is None:
            raise AssertionError("self.executor_instance cannot be None")

        self.executor_instance.run(
            ["apt-get", "update"], run_as_root=True, check=True
        )
        self.executor_instance.run(
            ["apt-get", "--yes", "install", "blhc"],
            run_as_root=True,
            check=True,
        )

        build_files = utils.find_files_suffixes(download_directory, [".build"])

        if len(build_files) == 0:
            list_of_files = sorted(map(str, download_directory.iterdir()))
            self.append_to_log_file(
                "configure_for_execution.log",
                [f"No *.build files to be analyzed. Files: {list_of_files}"],
            )
            return False

        self._blhc_target = build_files[0]

        return True

    @override
    def upload_artifacts(
        self,
        exec_directory: Path,
        *,
        execution_result: WorkRequestResults,  # noqa: U100
    ) -> None:
        """Upload the BlhcArtifact with the files and relationships."""
        assert self.dynamic_data
        if not self.debusine:
            raise AssertionError("self.debusine not set")

        blhc_file = exec_directory / self.CAPTURE_OUTPUT_FILENAME

        blhc_artifact = BlhcArtifact.create(
            blhc_output=blhc_file,
            build_log=self.dynamic_data.input_artifact_data,
        )

        uploaded = self.debusine.upload_artifact(
            blhc_artifact,
            workspace=self.workspace_name,
            work_request=self.work_request_id,
        )

        for source_artifact_id in self._source_artifacts_ids:
            self.debusine.relation_create(
                uploaded.id, source_artifact_id, RelationType.RELATES_TO
            )

    @override
    def task_result(
        self,
        returncode: int | None,
        execute_directory: Path,  # noqa: U100
    ) -> WorkRequestResults:
        """
        Evaluate task output and return success.

        We don't actually check the output, but use the return code of
        blhc.
        """
        if returncode in [0, 1]:
            return WorkRequestResults.SUCCESS
        return WorkRequestResults.FAILURE



"""
The settings in this file are tailored overrides for running in production.

When running in production, selected.py should point to this file.
"""

# PostgreSQL should be used in production
try:
    from debusine.project.settings.db_postgresql import DATABASES  # noqa: F401
except (ModuleNotFoundError, PermissionError):
    from debusine.project.settings.db_postgresql_default import (  # noqa: F401
        DATABASES,
    )

# Use paths from the package
from debusine.project.project_utils import read_secret_key
from debusine.project.settings.pkg_paths import *  # noqa: F401, F403

# Read the SECRET_KEY from the file
SECRET_KEY = read_secret_key(f"{DEBUSINE_DATA_PATH}/key")  # noqa: F405



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""debusine-client interactive data entry."""

import abc
import cmd
import inspect
import re
import readline
from typing import Any, Literal, overload, override

import rich.markup
from rich.console import Console


class DataEntry(cmd.Cmd, abc.ABC):
    """Base class for interactive data entry loops."""

    def __init__(self, console: Console | None = None) -> None:
        """Set up a Rich console for output."""
        super().__init__()
        self.console = console or Console()

    @overload
    def input_line(
        self,
        prompt: str,
        *,
        initial: str | None = None,
        required: Literal[True],
        validator: re.Pattern[str] | None = None,
    ) -> str: ...

    @overload
    def input_line(
        self,
        prompt: str,
        *,
        initial: str | None = None,
        required: Literal[False] = False,
        validator: re.Pattern[str] | None = None,
    ) -> str: ...

    def input_line(
        self,
        prompt: str,
        *,
        initial: str | None = None,
        required: bool = False,
        validator: re.Pattern[str] | None = None,
    ) -> str | None:
        """
        Input a line of text.

        :param prompt: prompt to display
        :param initial: initial value that the user can correct
        :param required: set to True to insist that a value is provided
        """
        if initial is not None:
            readline.set_startup_hook(
                lambda: readline.insert_text(initial)  # pragma: no cover
            )

        try:
            while True:
                # Rich plays badly with readline, see:
                # https://github.com/Textualize/rich/issues/2293
                self.console.print(prompt)
                result = input("> ")
                result = result.strip()
                if not result:
                    if required:
                        self.console.print(
                            "Please enter a value: this field is needed"
                        )
                    else:
                        return None
                elif validator and not validator.match(result):
                    self.console.print(
                        rich.markup.escape(repr(result)),
                        "should match"
                        f" [bold]{rich.markup.escape(validator.pattern)}[/]",
                    )
                else:
                    return result
        finally:
            readline.set_startup_hook()

    @override
    def do_help(self, arg: str) -> None:
        """
        Command line help.

        You can use `help command` for details about the command
        """
        arg = arg.strip().lower()
        self.console.print()
        if not arg:
            self.console.print(
                "Command summary (using the first letter is enough):"
            )
            self.console.print()

            for name, value in inspect.getmembers(self):
                if not name.startswith("do_"):
                    continue
                summary = (inspect.getdoc(value) or "").splitlines()[0].strip()
                self.console.print(f" * {name[3:]}: {summary}")

            self.console.print()
            self.console.print(
                r"You can use '[bold]h[/]elp' command for detailed help"
            )
        else:
            func = getattr(self, f"do_{arg}", None)
            if func is None:
                self.console.print(f"No help found for {arg!r}", markup=False)
            else:
                self.console.print(inspect.getdoc(func))
                self.console.print()

    def do_quit(self, arg: str) -> bool:  # noqa: ARG002, U100
        """Quit editor."""
        return True

    def do_eof(self, arg: str) -> bool:
        """Stop processing."""
        self.console.print()
        return self.do_quit(arg)

    @abc.abstractmethod
    def menu(self) -> None:
        """Print a menu for the user."""

    # Cmd.default is really expected to return a bool|None, and it looks like
    # badly typed (see how Cmd.onecmd uses it)
    @override
    def default(self, line: str) -> bool:  # type: ignore[override]
        """Print error to console."""
        self.console.print(
            f"[red]command not recognized: {rich.markup.escape(repr(line))}[/]",
            highlight=False,
        )
        return False

    @override
    def precmd(self, line: str) -> str:
        """Allow abbreviated commands."""
        parts = line.split(None, 1)
        if parts:
            for name, _value in inspect.getmembers(self):
                if name.startswith(f"do_{parts[0].lower()}"):
                    parts[0] = name[3:]
        return " ".join(parts)

    @override
    def postcmd(self, stop: bool, line: str) -> bool:
        """Show the menu before a prompt."""
        res = super().postcmd(stop, line)
        if not res:
            self.menu()
        return res

    @override
    def cmdloop(self, intro: Any | None = None) -> None:  # noqa: ARG002, U100
        """Show menu when entering the loop."""
        self.menu()
        super().cmdloop()



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Task to use autopkgtest in debusine."""

import logging
import re
import shlex
from pathlib import Path
from typing import Any, cast, override

import pydantic

import debusine.tasks.tags as ttags
import debusine.worker.tags as wtags
from debusine import utils
from debusine.artifacts.local_artifact import AutopkgtestArtifact
from debusine.artifacts.models import (
    ArtifactCategory,
    DebianAutopkgtest,
    DebianAutopkgtestResult,
    DebianAutopkgtestResultStatus,
    DebianAutopkgtestSource,
    DebianSourcePackage,
    DebianSystemImage,
    DebianSystemTarball,
    DebianUpload,
    WorkRequestResults,
    get_source_package_name,
    get_source_package_version,
)
from debusine.client.models import RelationType
from debusine.tasks import (
    BaseTaskWithExecutor,
    ExtraRepositoryMixin,
    RunCommandTask,
)
from debusine.tasks.inputs import (
    EnvironmentInput,
    MultiInput,
    SingleInput,
    Stage,
)
from debusine.tasks.models import (
    AutopkgtestData,
    AutopkgtestDynamicData,
    BackendType,
)

log = logging.getLogger(__name__)

type ParsedSummaryFile = dict[str, DebianAutopkgtestResult]


class Autopkgtest(
    ExtraRepositoryMixin[AutopkgtestData, AutopkgtestDynamicData],
    RunCommandTask[AutopkgtestData, AutopkgtestDynamicData],
    BaseTaskWithExecutor[AutopkgtestData, AutopkgtestDynamicData],
):
    """Task to use autopkgtest in debusine."""

    TASK_VERSION = 1

    ARTIFACT_DIR = "artifact-dir"
    SUMMARY_FILE = "artifact-dir/summary"

    input_source_artifact = SingleInput(
        field="input.source_artifact",
        categories=[ArtifactCategory.SOURCE_PACKAGE, ArtifactCategory.UPLOAD],
        stage=Stage.PENDING,
    )
    input_binary_artifacts = MultiInput(
        field="input.binary_artifacts",
        categories=[ArtifactCategory.BINARY_PACKAGE, ArtifactCategory.UPLOAD],
    )
    input_context_artifacts = MultiInput(
        field="input.context_artifacts",
        categories=[ArtifactCategory.BINARY_PACKAGE, ArtifactCategory.UPLOAD],
    )
    environment = EnvironmentInput(
        categories=[
            ArtifactCategory.SYSTEM_TARBALL,
            ArtifactCategory.SYSTEM_IMAGE,
        ],
        stage=Stage.PENDING,
    )

    @override
    def post_init(self) -> None:
        super().post_init()

        self._source_package_name: str | None = None
        self._source_package_version: str | None = None
        self._source_package_url: pydantic.AnyUrl | None = None
        self._source_package_path: str | None = None

        self._parsed: ParsedSummaryFile | None = None
        self._autopkgtest_targets: list[Path] = []

    @override
    @classmethod
    def analyze_worker(cls) -> dict[str, Any]:
        """Report metadata for this task on this worker."""
        metadata = super().analyze_worker()

        available_key = cls.prefix_with_task_name("available")
        metadata[available_key] = utils.is_command_available("autopkgtest")

        return metadata

    @override
    @classmethod
    def get_provided_worker_tags(cls) -> set[str]:
        tags = super().get_provided_worker_tags()
        if utils.is_command_available("autopkgtest"):
            tags.add(wtags.CAP_PREFIX + "autopkgtest")
        return tags

    @override
    def compute_scheduler_tags(
        self,
    ) -> tuple[ttags.TaskProvidedTags, wtags.TaskRequiredTags]:
        provided, required = super().compute_scheduler_tags()

        assert isinstance(
            self.input_source_artifact.data, (DebianSourcePackage, DebianUpload)
        )
        provided.add(
            ttags.ProvenanceProvided.SYSTEM,
            {
                ttags.SOURCE_PACKAGE_PREFIX
                + get_source_package_name(self.input_source_artifact.data)
            },
        )
        required.add(
            wtags.ProvenanceRequired.SYSTEM,
            {wtags.CAP_PREFIX + "autopkgtest"},
        )
        return provided, required

    @override
    def can_run_on(self, worker_metadata: dict[str, Any]) -> bool:
        """Check if the specified worker can run the task."""
        if not super().can_run_on(worker_metadata):
            return False

        executor_available_key = f"executor:{self.backend}:available"
        available_key = self.prefix_with_task_name("available")
        return bool(
            worker_metadata.get(executor_available_key, False)
            and worker_metadata.get(available_key, False)
        )

    @override
    def get_subject(self) -> str | None:
        assert isinstance(
            self.input_source_artifact.data, (DebianSourcePackage, DebianUpload)
        )
        return get_source_package_name(self.input_source_artifact.data)

    @override
    def get_configuration_context(self) -> str | None:
        assert isinstance(
            self.environment.data, (DebianSystemTarball, DebianSystemImage)
        )
        return self.environment.data.codename

    @override
    def compute_dynamic_data(self) -> AutopkgtestDynamicData:
        """
        Resolve artifact lookups for this task.

        :subject: source package name out of ``input:source_artifact``
        :parameter_summary: ``source_package_name_version (architecture)``
        :runtime_context: ``build_architecture:executor_backend``
        :configuration_context: codename of ``environment``
        """
        input_source_artifact = self.input_source_artifact
        assert isinstance(
            input_source_artifact.data, (DebianSourcePackage, DebianUpload)
        )

        package_name = get_source_package_name(input_source_artifact.data)
        package_version = get_source_package_version(input_source_artifact.data)

        return AutopkgtestDynamicData(
            environment_id=self.environment.artifact_id,
            extra_repositories=self.extra_repositories,
            input_source_artifact_id=input_source_artifact.artifact_id,
            input_binary_artifacts_ids=self.input_binary_artifacts.resolved_ids(),
            input_context_artifacts_ids=self.input_context_artifacts.resolved_ids(),
            subject=self.get_subject(),
            # TODO: It would probably be better to get the architecture of
            # the input binary artifacts and use that in the parameter
            # summary.
            parameter_summary=f"{package_name}_{package_version} "
            f"({self.build_architecture})",
            runtime_context=f"{self.build_architecture}:{self.backend}",
            configuration_context=self.get_configuration_context(),
        )

    @override
    def check_directory_for_consistency_errors(
        self, build_directory: Path
    ) -> list[str]:
        """Autopkgtest ARTIFACT_DIR/summary file does not exist."""
        summary_path = build_directory / self.SUMMARY_FILE
        if not summary_path.exists():
            return [f"'{self.SUMMARY_FILE}' does not exist"]

        return []

    @staticmethod
    def _parse_summary_file(
        summary_file: Path,
    ) -> ParsedSummaryFile:
        """
        Parse autopkgtest summary file (from autopkgtest --summary).

        :param summary_file: file to parse.
        :return: dictionary with the result. Structure:
            .. code-block::
                {"test-name-1": {"status": "PASS"},
                 "test-name-2: {"status": "FAIL", "details": "partial"},
                }

            "status": always in the dictionary (PASS, FAIL, FLAKY, SKIP or
            any other status written by autopkgtest)
            "details": only in the dictionary if details are found
            The test-name may be "*" if it wasn't known.
            If there's a testbed failure, the dictionary will be empty.
        :raises: ValueError if a line cannot be parsed
        """  # noqa: RST301
        parsed = {}
        with summary_file.open() as file:
            for line in file.readlines():
                line = line.rstrip()

                m = re.match(r"(?P<error>[a-z ]+): (?P<details>.*)", line)
                if m is not None:
                    log.info(
                        "Autopkgtest error: %s: %s",
                        m.group("error"),
                        m.group("details"),
                    )
                    continue

                m = re.match(
                    r"(?P<name>\S+)\s+(?P<status>\S+)"
                    r"(?:\s+(?P<details>.*))?",
                    line,
                )

                if m is None:
                    raise ValueError(f"Failed to parse line: {line}")

                name = m.group("name")
                result = m.group("status")

                if result not in ("PASS", "FAIL", "SKIP", "FLAKY"):
                    raise ValueError(f"Line with unexpected result: {line}")

                parsed[name] = DebianAutopkgtestResult(
                    status=DebianAutopkgtestResultStatus(result)
                )

                if details := m.group("details"):
                    parsed[name].details = details

        return parsed

    def _environment_codename(self) -> str:
        """Return the codename of the environment we're testing on."""
        if self.executor is None:
            raise ValueError(
                "self.executor not set, call _prepare_executor() first"
            )
        return cast(str, self.executor.system_image.data["codename"])

    @override
    def fetch_input(self, destination: Path) -> bool:
        """Download the required artifacts."""
        assert self.dynamic_data

        artifact = self.fetch_artifact(
            self.dynamic_data.input_source_artifact_id, destination
        )

        for file_path, file_data in artifact.files.items():
            if file_path.endswith(".dsc"):
                self._source_package_url = file_data.url
                self._source_package_path = file_path
                break

        for artifact_id in self.dynamic_data.input_binary_artifacts_ids:
            self.fetch_artifact(artifact_id, destination)

        for artifact_id in self.dynamic_data.input_context_artifacts_ids:
            self.fetch_artifact(artifact_id, destination)

        self._prepare_executor()
        self.write_extra_repository_config(
            codename=self._environment_codename(),
            destination=destination,
        )

        return True

    @override
    def _cmdline(self) -> list[str]:
        """
        Return autopkgtest command line (idempotent).

        Use configuration of self.data.
        """
        if not self.executor:
            raise AssertionError(
                "self.executor not set - self._prepare_for_execution() "
                "must be called before _cmdline()"
            )

        cmd = [
            "autopkgtest",
            "--apt-upgrade",
            f"--output-dir={self.ARTIFACT_DIR}",
            f"--summary={self.SUMMARY_FILE}",
            "--no-built-binaries",
        ]

        cmd += [
            f"--test-name={include_test}"
            for include_test in self.data.include_tests
        ]

        cmd += [
            f"--skip-test={exclude_test}"
            for exclude_test in self.data.exclude_tests
        ]

        if debug_level := self.data.debug_level:
            cmd.append("-" + "d" * debug_level)

        cmd += self._cmdline_extra_repository()

        if self.data.use_packages_from_base_repository:
            release = self._environment_codename()
            cmd.append(f"--apt-default-release={release}")

        for release, pkglist in self.data.pin_packages.items():
            cmd.append(f"--pin-packages={release}={','.join(pkglist)}")

        for variable, value in self.data.extra_environment.items():
            cmd.append(f"--env={variable}={value}")

        cmd.append(f"--needs-internet={self.data.needs_internet}")

        if self.data.timeout is not None:
            for key, timeout in self.data.timeout.model_dump(
                by_alias=True
            ).items():
                if timeout is not None:
                    cmd.append(f"--timeout-{key}={timeout}")

        if (
            self.backend == BackendType.UNSHARE
            and Path("/etc/resolv.conf").is_file()
        ):
            # autopkgtest >= 5.31 copies /etc/resolv.conf from the host, but
            # earlier versions don't.  Force this, since we can't rely on
            # resolver configuration built into the system tarball.
            cmd.append("--copy=/etc/resolv.conf:/etc/resolv.conf")

        cmd.extend(map(str, self._autopkgtest_targets))

        cmd.append("--")
        cmd.append(self.executor.autopkgtest_virt_server())
        cmd.extend(self.executor.autopkgtest_virt_args())

        return cmd

    def _cmdline_extra_repository(self) -> list[str]:
        """Generate command line arguments for extra_repositories."""
        args = [
            f"--copy={path}:/etc/apt/sources.list.d/{path.name}"
            for path in self.extra_repository_sources
        ]
        # autopkgtest has native deb822 support coming soon: #1089736
        # apt < 2.3.10 has no support for keys embedded in Signed-By
        args += [
            f"--copy={path}:/etc/apt/keyrings/{path.name}"
            for path in self.extra_repository_keys
        ]
        return args

    @override
    def configure_for_execution(self, download_directory: Path) -> bool:
        """Gather information used later on (_cmdline(), upload_artifacts())."""
        # Not yet strictly guaranteed, because the source artifact might be
        # of the wrong type; see
        # https://salsa.debian.org/freexian-team/debusine/-/issues/207.
        assert self._source_package_path is not None

        # Used by upload_artifacts()
        dsc_file = download_directory / self._source_package_path
        try:
            dsc = utils.read_dsc(dsc_file)
        except ValueError as e:
            self.append_to_log_file("configure_for_execution.log", [str(e)])
            return False

        self._source_package_name = dsc["source"]
        self._source_package_version = dsc["version"]

        # Used by _cmdline():
        self._autopkgtest_targets = utils.find_files_suffixes(
            download_directory, [".deb"]
        )
        self._autopkgtest_targets.append(dsc_file)

        return True

    @override
    def task_result(
        self, returncode: int | None, execute_directory: Path
    ) -> WorkRequestResults:
        """
        Parse the summary file and return success.

        Use self.data.fail_on.
        """
        self._parsed = self._parse_summary_file(
            execute_directory / self.SUMMARY_FILE
        )

        fail_on = self.data.fail_on

        for result in self._parsed.values():
            if (
                (result.status == "FAIL" and fail_on.failed_test)
                or (result.status == "FLAKY" and fail_on.flaky_test)
                or (result.status == "SKIP" and fail_on.skipped_test)
            ):
                return WorkRequestResults.FAILURE

        if returncode == 4 and not fail_on.failed_test:
            return WorkRequestResults.SUCCESS

        if returncode == 2 and fail_on.flaky_test:
            return WorkRequestResults.FAILURE

        if returncode == 8 and fail_on.skipped_test:
            return WorkRequestResults.FAILURE

        if returncode == 8 and all(
            result.status == "SKIP" for result in self._parsed.values()
        ):
            return WorkRequestResults.SKIPPED

        # Return True if autopkgtest has run successfully.
        # 0 all tests passed
        # 2 at least one test was skipped (or at least one flaky test failed)
        # 8 no tests in this package, or all non-superficial tests were skipped
        if returncode in {0, 2, 8}:
            return WorkRequestResults.SUCCESS
        return WorkRequestResults.FAILURE

    @override
    def upload_artifacts(
        self,
        execute_directory: Path,
        *,
        execution_result: WorkRequestResults,  # noqa: U100
    ) -> None:
        """Upload AutopkgtestArtifact with the files, data and relationships."""
        if not self.debusine:
            raise AssertionError("self.debusine not set")

        if not self.executor:
            raise AssertionError(
                "self.executor not set - self._prepare_for_execution() "
                "must be called before upload_artifacts()"
            )

        img_data = self.executor.system_image.data

        assert self._parsed is not None
        assert self._source_package_path is not None
        assert self._source_package_name is not None
        assert self._source_package_version is not None
        assert self._source_package_url is not None

        autopkgtest_artifact = AutopkgtestArtifact.create(
            execute_directory / self.ARTIFACT_DIR,
            DebianAutopkgtest(
                results=self._parsed,
                cmdline=shlex.join(self._cmdline()),
                source_package=DebianAutopkgtestSource(
                    name=self._source_package_name,
                    version=self._source_package_version,
                    url=self._source_package_url,
                ),
                # TODO: Should this be the architecture of the binary
                # packages instead?  If we're testing `Architecture: all`
                # binary packages on amd64, "all" seems like a better value
                # here.
                architecture=self.data.build_architecture,
                distribution=f"{img_data['vendor']}:{img_data['codename']}",
            ),
        )

        uploaded = self.debusine.upload_artifact(
            autopkgtest_artifact,
            workspace=self.workspace_name,
            work_request=self.work_request_id,
        )

        assert self.dynamic_data
        self.debusine.relation_create(
            uploaded.id,
            self.dynamic_data.input_source_artifact_id,
            RelationType.RELATES_TO,
        )

        for artifact_id in self.dynamic_data.input_binary_artifacts_ids:
            self.debusine.relation_create(
                uploaded.id,
                artifact_id,
                RelationType.RELATES_TO,
            )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the ImageCache class."""

import datetime as dt
import json
from pathlib import Path
from shutil import rmtree
from tempfile import mkdtemp
from typing import override
from unittest.mock import MagicMock, create_autospec, patch

from debusine.artifacts.models import ArtifactCategory
from debusine.client.debusine import Debusine
from debusine.tasks import Noop
from debusine.tasks.executors.base import (
    ExecutorImageCategory,
    ExecutorInterface,
    _backends,
)
from debusine.tasks.executors.images import CACHE_SIZE, ImageCache
from debusine.tasks.executors.images import log as images_log
from debusine.tasks.models import BackendType
from debusine.tasks.tests.helpers import ExternalTaskHelperMixin
from debusine.test import TestCase


class BaseImageCacheTests(ExternalTaskHelperMixin[Noop], TestCase):
    """Common setup for unit tests of ImageCache class."""

    @override
    def setUp(self) -> None:
        """Mock the Debusine API for tests."""
        super().setUp()
        self.debusine_api = create_autospec(Debusine)
        self.image_cache = ImageCache(
            self.debusine_api,
            ExecutorImageCategory.TARBALL,
            workspace="workspace",
        )
        self.artifact = self.fake_system_tarball_artifact()
        self.debusine_api.artifact_get.return_value = self.artifact


class ImageCacheMetadataTests(BaseImageCacheTests):
    """Unit tests for ImageCache class that don't work with images."""

    def test_cache_artifact_path(self) -> None:
        """Test that cache_artifact_path builds the correct path."""
        self.assertEqual(
            self.image_cache.cache_artifact_path(42),
            self.image_cache.image_cache_path / "42",
        )

    def test_cache_artifact_image_path(self) -> None:
        """Test that cache_artifact_image_path builds the correct base path."""
        self.assertEqual(
            self.image_cache.cache_artifact_image_path(42, "foo.img"),
            self.image_cache.image_cache_path / "42" / "foo.img",
        )

    def test_cache_artifact_image_path_for_backend(self) -> None:
        """Test that cache_artifact_image_path builds a correct backend path."""
        self.assertEqual(
            self.image_cache.cache_artifact_image_path(
                42, "foo.img", backend=BackendType.UNSHARE
            ),
            self.image_cache.image_cache_path
            / "42"
            / "backends"
            / "unshare"
            / "foo.img",
        )

    def test_image_artifact(self) -> None:
        """Test that image_metadata fetches the Artifact."""
        self.assertEqual(self.image_cache.image_artifact(42), self.artifact)

        self.debusine_api.artifact_get.assert_called_once_with(
            42, workspace="workspace"
        )

    def test_image_artifact_incorrect_category(self) -> None:
        """Test that image_artifact rejects the wrong kind of artifact."""
        self.artifact.category = ArtifactCategory.SYSTEM_IMAGE

        with self.assertRaisesRegex(
            ValueError,
            r"^Unexpected artifact type debian:system-image; expected "
            r"debian:system-tarball$",
        ):
            self.image_cache.image_artifact(42)


class ImageCacheTests(BaseImageCacheTests):
    """Unit tests for ImageCache class that work with images."""

    @override
    def setUp(self) -> None:
        """Mock the cache directory for tests."""
        super().setUp()
        self.cache_path = Path(mkdtemp(prefix="debusine-testsuite-images-"))
        self.addCleanup(rmtree, self.cache_path)
        self.image_cache.image_cache_path = self.cache_path
        patcher = patch.object(self.image_cache, 'process_downloaded_image')
        self.process_downloaded_image_mock = patcher.start()
        self.addCleanup(patcher.stop)

    def patch_download_image_artifact(self) -> None:
        """Mock ImageCache._download_image_artifact()."""
        patcher = patch.object(self.image_cache, "_download_image_artifact")
        self._download_image_artifact_mock = patcher.start()
        self.addCleanup(patcher.stop)

    def test_download_image_uncached(self) -> None:
        """Test that download_image downloads images when not cached."""
        self.patch_download_image_artifact()
        image = self.image_cache.download_image(self.artifact)

        self.assertEqual(image, self.cache_path / "42" / "system.tar.xz")
        self._download_image_artifact_mock.assert_called_with(
            42, "system.tar.xz", self.cache_path / "42" / "system.tar.xz"
        )
        self.process_downloaded_image_mock.assert_called_with(self.artifact)

    def test_download_image_cached(self) -> None:
        """Test that download_image uses cached images."""
        self.patch_download_image_artifact()
        image_path = self.cache_path / "42" / "system.tar.xz"
        image_path.parent.mkdir(parents=True)
        image_path.write_bytes(b"Fake Data")
        self.assertEqual(
            self.image_cache.download_image(self.artifact), image_path
        )
        self._download_image_artifact_mock.assert_not_called()
        self.process_downloaded_image_mock.assert_not_called()

    def test_download_image_cached_artifact(self) -> None:
        """Test that download_image processes cached artifacts."""
        self.patch_download_image_artifact()

        # The image still needs processing:
        patcher = patch.object(self.image_cache, "image_in_cache")
        image_in_cache_mock = patcher.start()
        self.addCleanup(patcher.stop)
        image_in_cache_mock.return_value = False

        image_path = self.cache_path / "42" / "system.tar.xz"
        image_path.parent.mkdir(parents=True)
        image_path.write_bytes(b"Fake Data")

        self.image_cache.download_image(self.artifact)

        self._download_image_artifact_mock.assert_not_called()
        self.process_downloaded_image_mock.assert_called_with(self.artifact)

    def test_download_image_invalid_name(self) -> None:
        """Test that download_image downloads images when not cached."""
        self.patch_download_image_artifact()
        self.artifact.data["filename"] = ".hidden.artifact"
        with self.assertRaisesRegex(
            ValueError, "Image has an invalid filename"
        ):
            self.image_cache.download_image(self.artifact)

    def test_image_in_cache_absent(self) -> None:
        """Test that image_in_cache returns false when not cached."""
        self.assertFalse(self.image_cache.image_in_cache(self.artifact))

    def test_image_in_cache_present(self) -> None:
        """Test that image_in_cache finds images in the cache."""
        image_path = self.cache_path / "42" / "system.tar.xz"
        image_path.parent.mkdir(parents=True)
        image_path.write_bytes(b"Fake Data")
        self.assertTrue(self.image_cache.image_in_cache(self.artifact))

    def test_download_image_artifact(self) -> None:
        """Test that _download_image_artifact does a 2-stage download."""

        def download_tarball(
            artifact_id: int,  # noqa: ARG001, U100
            path_in_artifact: str,  # noqa: ARG001, U100
            destination: Path,
            workspace: str,  # noqa: ARG001, U100
        ) -> None:
            destination.write_bytes(b"downloaded!")

        self.debusine_api.download_artifact_file.side_effect = download_tarball

        destination = self.cache_path / "filename.tar"
        self.image_cache._download_image_artifact(
            42, "filename.tar", destination
        )

        temp_path = self.cache_path / ".download.tmp.filename.tar"
        self.debusine_api.download_artifact_file.assert_called_with(
            42,
            path_in_artifact="filename.tar",
            destination=temp_path,
            workspace="workspace",
        )

        self.assertTrue(destination.exists())

    @patch("debusine.tasks.executors.images.now")
    def test_record_image_use(self, now_mock: MagicMock) -> None:
        """Test that record_image_use writes a usage log."""
        (dir_ := self.image_cache.cache_artifact_path(42)).mkdir()
        record_path = dir_ / "usage.json"
        now_mock.return_value = dt.datetime(
            2024, 1, 1, 0, 0, 0, 0, tzinfo=dt.UTC
        )

        self.image_cache.record_image_use(self.artifact)

        self.assertTrue(record_path.exists())
        with record_path.open("r") as f:
            record = json.load(f)
        self.assertEqual(
            record,
            {
                "version": 1,
                "backends": [],
                "usage": [
                    {
                        "filename": "system.tar.xz",
                        "timestamp": "2024-01-01T00:00:00Z",
                        "backend": None,
                    }
                ],
            },
        )

    @patch("debusine.tasks.executors.images.now")
    def test_record_image_use_appends(self, now_mock: MagicMock) -> None:
        """Test that record_image_use updates an existing usage log."""
        (dir_ := self.image_cache.cache_artifact_path(42)).mkdir()
        record_path = dir_ / "usage.json"
        existing = {
            "version": 1,
            "backends": [],
            "usage": [
                {
                    "filename": "system.tar.xz",
                    "timestamp": "2024-01-01T00:00:00Z",
                    "backend": None,
                }
            ]
            * 10,
        }
        with record_path.open("w") as f:
            json.dump(existing, f)

        now_mock.return_value = dt.datetime(
            2024, 1, 2, 0, 0, 0, 0, tzinfo=dt.UTC
        )

        self.image_cache.record_image_use(self.artifact)

        with record_path.open("r") as f:
            record = json.load(f)

        self.assertEqual(len(record["usage"]), 10)
        self.assertEqual(
            record["usage"][-1],
            {
                "filename": "system.tar.xz",
                "timestamp": "2024-01-02T00:00:00Z",
                "backend": None,
            },
        )

    def test_find_least_used_images_empty_cache(self) -> None:
        """Test _find_least_used_images with an empty cache."""
        self.assertEqual(self.image_cache._find_least_used_images(), [])

    def test_find_least_used_images_ignores_non_digit_directories(self) -> None:
        """Test _find_least_used_images ignores non-digit directories."""
        (self.image_cache.image_cache_path / "foo").mkdir()
        self.assertEqual(self.image_cache._find_least_used_images(), [])

    def test_find_least_used_images_missing_log(self) -> None:
        """Test _find_least_used_images with an image missing a log."""
        self.image_cache.cache_artifact_path(42).mkdir()
        with self.assertLogsContains(
            "No usage log for artifact 42", logger=images_log
        ):
            self.assertEqual(
                self.image_cache._find_least_used_images(),
                [(42, set(_backends.keys()))],
            )

    def test_find_least_used_images_invalid_log(self) -> None:
        """Test _find_least_used_images with an unparsable log."""
        (cache_dir := self.image_cache.cache_artifact_path(42)).mkdir()
        with (cache_dir / "usage.json").open("w") as f:
            f.write('{"version": 99}')
        with self.assertLogsContains(
            "Failed to parse log for artifact 42", logger=images_log
        ):
            self.assertEqual(
                self.image_cache._find_least_used_images(),
                [(42, set(_backends.keys()))],
            )

    def write_image_usage_log(
        self,
        artifact_id: int,
        timestamp: dt.datetime | None = None,
        backends: set[BackendType] | None = None,
    ) -> None:
        """Create an image usage log for artifact_id, with timestamp."""
        if timestamp is None:
            timestamp = dt.datetime(2024, 1, 1, 0, 0, 0, 0, tzinfo=dt.UTC)
        if backends is None:
            backends = set()
        (dir_ := self.image_cache.cache_artifact_path(artifact_id)).mkdir()
        record_path = dir_ / "usage.json"
        existing = {
            "version": 1,
            "backends": list(backends),
            "usage": [
                {
                    "filename": "system.tar.xz",
                    "timestamp": timestamp.isoformat(),
                    "backend": None,
                }
            ],
        }
        with record_path.open("w") as f:
            json.dump(existing, f)

    def test_find_least_used_images_under_size(self) -> None:
        """Test _find_least_used_images returns nothing under CACHE_SIZE."""
        assert CACHE_SIZE > 2
        self.write_image_usage_log(42)
        self.write_image_usage_log(43)
        self.assertEqual(self.image_cache._find_least_used_images(), [])

    def test_find_least_used_images_over_size(self) -> None:
        """Test _find_least_used_images returns images over CACHE_SIZE."""
        for i in range(CACHE_SIZE):
            self.write_image_usage_log(i)
        self.write_image_usage_log(
            99,
            dt.datetime(2023, 12, 1, 0, 0, 0, 0, tzinfo=dt.UTC),
            {BackendType.UNSHARE},
        )
        self.assertEqual(
            self.image_cache._find_least_used_images(),
            [(99, {BackendType.UNSHARE})],
        )

    @patch("debusine.tasks.executors.images.rmtree")
    def test_clean_up_old_images(self, rmtree: MagicMock) -> None:
        """Test that clean_up_old_images cleans them up."""
        patcher = patch.object(self.image_cache, "_find_least_used_images")
        _find_least_used_images = patcher.start()
        self.addCleanup(patcher.stop)
        _find_least_used_images.return_value = [(99, {"foo"})]
        (img_dir := self.image_cache.cache_artifact_path(99)).mkdir()

        FooExecutor = MagicMock()

        with patch.dict(
            "debusine.tasks.executors.base._backends", {"foo": FooExecutor}
        ):
            self.image_cache.clean_up_old_images()

        FooExecutor.clean_up_image.assert_called_once_with(99)
        rmtree.assert_called_once_with(img_dir)

    @patch("debusine.tasks.executors.images.rmtree")
    def test_clean_up_old_images_ignores_backends_without_cleanup(
        self, rmtree: MagicMock
    ) -> None:
        """Test that clean_up_old_images doesn't require backends cleeanup."""
        patcher = patch.object(self.image_cache, "_find_least_used_images")
        _find_least_used_images = patcher.start()
        self.addCleanup(patcher.stop)
        _find_least_used_images.return_value = [(99, {"foo"})]
        (img_dir := self.image_cache.cache_artifact_path(99)).mkdir()

        FooExecutor = create_autospec(ExecutorInterface)

        with patch.dict(
            "debusine.tasks.executors.base._backends", {"foo": FooExecutor}
        ):
            self.image_cache.clean_up_old_images()

        self.assertFalse(hasattr(FooExecutor, "clean_up_image"))
        rmtree.assert_called_once_with(img_dir)

    def test_clean_up_old_images_swallows_exceptions(self) -> None:
        """Test that clean_up_old_images swallows exceptions for backends."""
        patcher = patch.object(self.image_cache, "_find_least_used_images")
        _find_least_used_images = patcher.start()
        self.addCleanup(patcher.stop)
        _find_least_used_images.return_value = [(99, {"foo"})]
        self.image_cache.cache_artifact_path(99).mkdir()

        FooExecutor = MagicMock()
        FooExecutor.clean_up_image.side_effect = Exception("Can't find it.")

        with (
            patch.dict(
                "debusine.tasks.executors.base._backends", {"foo": FooExecutor}
            ),
            self.assertLogsContains(
                "Failed to clean up 99 backend foo", logger=images_log
            ),
        ):
            self.image_cache.clean_up_old_images()



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""debusine-admin command to print server configuration information."""

import datetime as dt  # pragma: no cover
from typing import Any, NoReturn, override  # pragma: no cover

from django.conf import settings  # pragma: no cover

from debusine.db.models import WorkRequest, Worker  # pragma: no cover
from debusine.django.management.debusine_base_command import (
    DebusineBaseCommand,
)  # pragma: no cover
from debusine.server.scheduler import (
    get_provided_worker_tags,
    get_required_task_tags,
)  # pragma: no cover


def format_date(value: dt.datetime | None) -> str:  # pragma: no cover
    """Format a datetime that could possibly be None."""
    return "never" if value is None else value.strftime("%Y-%m-%d")


class Command(DebusineBaseCommand):  # pragma: no cover
    """Command to check support for scheduler tags."""

    help = (
        "Check to see possible issues before turning on tag-based scheduling."
    )

    def print_work_request(self, wr: WorkRequest) -> None:
        """Format a work request in one line."""
        pending = format_date(wr.pending_at)
        started = format_date(wr.started_at)
        self.stdout.write(
            f"#{wr.id} created:{wr.created_at:%Y-%m-%d}"
            f" pending:{pending} started:{started}"
            f" created_by:{wr.created_by.username} status:{wr.status}"
            f" type:{wr.task_type} name:{wr.task_name}"
        )

    def check_workers_without_tags(self) -> None:
        """Check workers without tags."""
        workers_without_tags = Worker.objects.exclude(
            static_metadata__has_key="provided_tags"
        ).exclude(dynamic_metadata__has_key="provided_tags")
        if workers_without_tags.exists():
            self.stdout.write(
                f"Found {workers_without_tags.count()} workers"
                " without worker-provided tags:"
            )
            for worker in workers_without_tags:
                connected = format_date(worker.connected_at)
                self.stdout.write(
                    f"{worker.name} registered:{worker.registered_at:%Y-%m-%d}"
                    f" connected:{connected}"
                )
        else:
            self.stdout.write("Found no workers without worker-provided tags.")

    def check_work_requests_without_tags(self) -> None:
        """Check work requests without tags."""
        work_requests_without_tags = WorkRequest.objects.filter(
            status__in=(
                WorkRequest.Statuses.PENDING,
                WorkRequest.Statuses.RUNNING,
            ),
            scheduler_tags_provided__len=0,
            scheduler_tags_required__len=0,
        ).select_related("created_by")
        if work_requests_without_tags.exists():
            self.stdout.write(
                f"Found {work_requests_without_tags.count()} pending/running"
                " work_requests without task-provided or worker-required tags:"
            )
            for wr in work_requests_without_tags:
                self.print_work_request(wr)
        else:
            self.stdout.write(
                "Found no work_requests without worker-provided tags."
            )

    def check_scheduled_work_requests(self) -> None:
        """Check that scheduled work requests can be scheduled with tags."""
        work_requests_assigned = (
            WorkRequest.objects.filter(status=WorkRequest.Statuses.RUNNING)
            .exclude(worker=None)
            .select_related("worker")
        )
        assigned: int = work_requests_assigned.count()
        unschedulable: list[WorkRequest] = []
        for wr in work_requests_assigned:
            assert wr.worker is not None
            provided_worker_tags = get_provided_worker_tags(wr.worker)
            required_task_tags = get_required_task_tags(wr.worker)
            if not (
                set(wr.scheduler_tags_required) <= provided_worker_tags
                and required_task_tags <= set(wr.scheduler_tags_provided)
            ):
                unschedulable.append(wr)
        if assigned and unschedulable:
            self.stdout.write(
                f"{len(unschedulable)} unschedulable work requests among the"
                f" {assigned} assigned to workers:"
            )
            for wr in unschedulable:
                self.print_work_request(wr)
        elif assigned:
            self.stdout.write(
                f"No unschedulable work requests among the {assigned}"
                " assigned to workers."
            )
        else:
            self.stdout.write("No work requests currently assigned to workers.")

    def check_unschedulable_work_requests(self) -> None:
        """Check that pending work requests can be scheduled by one worker."""
        pending_wrs = WorkRequest.objects.filter(
            status=WorkRequest.Statuses.PENDING
        )
        if pending_wrs.exists():
            schedulable_wrs: set[int] = set()
            for worker in Worker.objects.all():
                provided_worker_tags = get_provided_worker_tags(worker)
                required_task_tags = get_required_task_tags(worker)
                for wr in pending_wrs.filter(
                    scheduler_tags_required__contained_by=list(
                        provided_worker_tags
                    ),
                    scheduler_tags_provided__contains=list(required_task_tags),
                ):
                    schedulable_wrs.add(wr.id)
            unschedulable_wrs = pending_wrs.exclude(
                id__in=list(schedulable_wrs)
            )
            if unschedulable_wrs.exists():
                self.stdout.write(
                    f"{unschedulable_wrs.count()}/{pending_wrs.count()}"
                    " pending work requests cannot be scheduled based on tags:"
                )
                for wr in unschedulable_wrs:
                    self.print_work_request(wr)
            else:
                self.stdout.write(
                    "All pending work requests can be scheduled using tags"
                    " by at least one of the current workers."
                )
        else:
            self.stdout.write(
                "There are no pending work requests to check for workers"
                " available for them via tags."
            )

    @override
    def handle(self, *args: Any, **options: Any) -> NoReturn:
        self.check_workers_without_tags()
        self.stdout.write()
        self.check_work_requests_without_tags()
        self.stdout.write()
        self.check_scheduled_work_requests()
        self.stdout.write()
        self.check_unschedulable_work_requests()
        self.stdout.write()
        feature_status = (
            "enabled"
            if getattr(settings, "FEATURE_TAG_BASED_SCHEDULING", True)
            else "disabled"
        )
        self.stdout.write(f"Tag based scheduling is {feature_status}.")
        raise SystemExit(0)



# Generated by Django 4.2.21 on 2025-09-16 10:17

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0016_artifact_db_artifact_workspace_category'),
    ]

    operations = [
        migrations.AddConstraint(
            model_name='assetusage',
            constraint=models.UniqueConstraint(
                fields=('asset', 'workspace'),
                name='db_assetusage_unique_asset_workspace',
            ),
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""debusine list view paginators."""

from collections import defaultdict
from typing import Any, override

from django.db.models import Model, Q, QuerySet
from django_cte import CTE, with_cte

from debusine.artifacts.models import TaskTypes
from debusine.db.models import (
    Artifact,
    CollectionItem,
    User,
    WorkRequest,
    Worker,
    WorkerPool,
    WorkflowTemplate,
)
from debusine.db.models.auth import GroupAuditLog, GroupMembership
from debusine.db.models.work_requests import WorkRequestQuerySet
from debusine.web.forms import WorkflowFilterForm
from debusine.web.views.table import (
    Column,
    DateTimeColumn,
    FilterField,
    FilterSelectOne,
    FilterToggle,
    NumberColumn,
    Ordering,
    Paginator,
    StringColumn,
    Table,
)


class GroupMembershipTable(Table[GroupMembership]):
    """Table set up for GroupMembership."""

    name = StringColumn(
        "Name",
        ordering=Ordering(
            asc=["user__username", "role"], desc=["-user__username", "role"]
        ),
    )
    display_name = StringColumn("Display name")
    role = StringColumn(
        "Role",
        ordering=Ordering(
            asc=["role", "user__username"], desc=["-role", "user__username"]
        ),
    )

    template_name = "web/_group_membership-list.html"
    default_order = "role"
    related_fields = ("user",)


class GroupMembershipAdminTable(GroupMembershipTable):
    """GroupMembership table with an Actions column."""

    actions = Column("Actions")


class GroupAuditLogTable(Table[GroupAuditLog]):
    """Table set up for GroupAuditLog."""

    date = DateTimeColumn("Date", ordering="created_at")
    actor = StringColumn(
        "User",
        ordering=Ordering(
            asc=["actor__username", "-created_at"],
            desc=["-actor__username", "-created_at"],
        ),
    )
    message = StringColumn("Message")

    template_name = "web/_group_audit_log-list.html"
    default_order = "-date"


class WorkerTable(Table[Worker]):
    """Table set up for Worker."""

    type_ = StringColumn("Type", ordering="worker_type", name="type")
    name = StringColumn("Name", ordering="name")
    pool = StringColumn("Pool", ordering="worker_pool")
    last_seen = DateTimeColumn("Last seen", ordering="token__last_seen_at")
    status = StringColumn("Status")

    filter_name = FilterField("Name")
    filter_pool = FilterSelectOne("Worker pool")
    filter_status = FilterSelectOne("Status")

    template_name = "web/_worker-list.html"
    default_order = "name"

    @override
    def init(self) -> None:
        """Add filter options based on database contents."""
        super().init()

        # Add options to worker pool filter
        self.filters["pool"].add_option(
            "_none", "Static workers", Q(worker_pool__isnull=True)
        )
        for pk, name in (
            WorkerPool.objects.filter(worker__in=self.queryset)
            .order_by("name")
            .distinct()
            .values_list("id", "name")
        ):
            self.filters["pool"].add_option(
                name=str(pk), label=name, query=Q(worker_pool__id=pk)
            )

        # Add options to status filter
        connected = Worker.objects.connected()
        self.filters["status"].add_option(
            "disabled",
            "Disabled",
            Q(token__isnull=True) | Q(token__enabled=False),
        )
        self.filters["status"].add_option(
            "disconnected",
            "Disconnected",
            Q(token__enabled=True) & ~Q(pk__in=connected),
        )
        self.filters["status"].add_option(
            "running",
            "Running",
            Q(
                token__enabled=True,
                pk__in=connected,
                assigned_work_requests__status=(WorkRequest.Statuses.RUNNING),
            ),
        )
        self.filters["status"].add_option(
            "idle",
            "Idle",
            Q(token__enabled=True, pk__in=connected)
            & ~Q(
                assigned_work_requests__status=(WorkRequest.Statuses.RUNNING),
            ),
        )


class WorkerPoolTable(Table[WorkerPool]):
    """Table set up for WorkerPool."""

    name = StringColumn("Name", ordering="name")
    enabled = StringColumn("Enabled", ordering="enabled")
    instance_wide = StringColumn("Instance-wide", ordering="instance_wide")
    ephemeral = StringColumn("Ephemeral", ordering="ephemeral")
    workers = StringColumn("Workers", ordering="workers")

    template_name = "web/_worker_pool-list.html"
    default_order = "name"


class WorkRequestTable(Table[WorkRequest]):
    """Table set up for WorkRequest."""

    id_ = NumberColumn("ID", ordering="pk", name="id")
    created_at = DateTimeColumn("Created", ordering="created_at")
    task_type = StringColumn("Task type", ordering="task_type")
    task_name = StringColumn("Task", ordering="task_name")
    status = StringColumn("Status", ordering="status")
    result = StringColumn("Result", ordering="result")

    template_name = "web/_work_request-list.html"
    default_order = "-created_at"
    related_fields = (
        "workspace__scope",
        "aborted_by",
    )


class FilterStatuses(FilterField):
    """Custom queryset filtering for workflow statuses."""

    @override
    def filter_queryset[M: Model](
        self, queryset: QuerySet[M, M], value: Any
    ) -> QuerySet[M, M]:
        assert isinstance(value, list)
        q_objects = Q()
        for status in value:
            if status == "running__any":
                q_objects |= Q(status=WorkRequest.Statuses.RUNNING)
            elif status.startswith("running__"):
                q_objects |= Q(
                    status=WorkRequest.Statuses.RUNNING,
                    workflow_runtime_status=status[9:],
                )
            else:
                q_objects |= Q(status=status)
        return queryset.filter(q_objects)


class FilterFailedWorkRequests(FilterToggle):
    """Keep only workflows with failed work requests."""

    def __init__(self, label: str) -> None:
        """Pass an empty Q, since we ignore it for filtering."""
        super().__init__(label, q=Q())

    @override
    def filter_queryset[M: Model](
        self, queryset: QuerySet[M, M], value: Any
    ) -> QuerySet[M, M]:
        if not value:
            return queryset
        assert isinstance(queryset, WorkRequestQuerySet)

        # Since we're comparing id to workflow_root_id (which is much more
        # efficient than having to compute the ancestor tree on the fly),
        # this can only be used to filter top-level workflows.
        assert not queryset.filter(parent__isnull=False).exists()

        workflows = CTE(queryset, name="workflows")
        filtered = with_cte(
            workflows,
            select=workflows.join(WorkRequest, id=workflows.col.id).filter(
                id__in=(
                    workflows.join(
                        WorkRequest, workflow_root_id=workflows.col.id
                    )
                    .exclude(task_type=TaskTypes.WORKFLOW)
                    .exclude(
                        workflow_data_json__contains={"allow_failure": True}
                    )
                    .visible_in_workflow()
                    .filter(result=WorkRequest.Results.FAILURE)
                    .values("workflow_root_id")
                    .order_by("workflow_root_id")
                    .distinct("workflow_root_id")
                )
            ),
        )
        assert isinstance(filtered, QuerySet)
        return filtered


class WorkflowPaginator(Paginator[WorkRequest]):
    """Paginator for workflow WorkRequests."""

    @override
    def get_context_data(self) -> dict[str, Any]:
        """Return context data for this paginator."""
        ctx = super().get_context_data()

        object_list = self.page_obj.object_list
        assert isinstance(object_list, WorkRequestQuerySet)
        workflows_flattened: dict[int, list[WorkRequest]] = defaultdict(list)
        work_request_status_counts: dict[
            int, dict[WorkRequest.Statuses, int]
        ] = defaultdict(lambda: defaultdict(int))
        work_request_result_counts: dict[
            int, dict[WorkRequest.Results, int]
        ] = defaultdict(lambda: defaultdict(int))
        if object_list.exists():
            for work_request in object_list.workflow_flattened():
                workflows_flattened[work_request.ancestor_id].append(
                    work_request
                )
                work_request_status_counts[work_request.ancestor_id][
                    work_request.status
                ] += 1
                if work_request.result:
                    # Work requests that fail but are allowed
                    # to fail intentionally disappear from
                    # the result counts. Other approaches
                    # (like counting as success, counting as
                    # failure with tricky UI coloring, or
                    # adding new badge types) added too much
                    # confusion or UI noise.
                    if (
                        work_request.result == WorkRequest.Results.FAILURE
                        and getattr(
                            work_request.workflow_data, "allow_failure", False
                        )
                    ):
                        continue
                    work_request_result_counts[work_request.ancestor_id][
                        work_request.result
                    ] += 1
        ctx["workflows_flattened"] = workflows_flattened
        ctx["work_request_status_counts"] = work_request_status_counts
        ctx["work_request_result_counts"] = work_request_result_counts

        return ctx


class WorkflowTable(Table[WorkRequest]):
    """Table set up for workflow WorkRequests."""

    id_ = NumberColumn("ID", ordering="pk", name="id")
    workflow_template = StringColumn("Template", ordering="task_name")
    status = StringColumn(
        "Status", ordering=("status", "workflow_runtime_status")
    )
    result = StringColumn("Result", ordering="result")
    started_at = DateTimeColumn("Started", ordering="started_at")
    completed_at = DateTimeColumn("Completed", ordering="completed_at")
    wr_count = Column("Count of work requests")
    last_activity = DateTimeColumn(
        "Last activity",
        ordering="workflow_last_activity_at",
        help_text="Latest activity recorded for workflow’s work request",
    )
    started_by = StringColumn("Started by", ordering="created_by__username")

    filter_statuses = FilterStatuses()
    filter_results = FilterField(q_field="result__in")
    filter_workflow_templates = FilterSelectOne("Workflow template")
    filter_started_by = FilterSelectOne("Started by")
    filter_with_failed_work_requests = FilterFailedWorkRequests(
        "With failed work requests"
    )

    template_name = "web/_workflow-list.html"
    default_order = "id.desc"
    filter_form_class = WorkflowFilterForm
    related_fields = ("workspace__scope", "created_by")

    @override
    def init_filters(self) -> None:
        """Add filter options based on database contents."""
        super().init_filters()
        for wt in (
            WorkflowTemplate.objects.filter(
                pk__in=(
                    self.queryset.values("workflow_template")
                    .order_by("workflow_template")
                    .distinct("workflow_template")
                )
            )
            .order_by("name")
            .distinct()
        ):
            self.filters["workflow_templates"].add_option(
                wt.name, wt.name, query=Q(workflow_template=wt)
            )

        for user in (
            User.objects.filter(
                pk__in=(
                    self.queryset.values("created_by")
                    .order_by("created_by")
                    .distinct("created_by")
                )
            )
            .order_by("first_name", "last_name")
            .distinct()
        ):
            self.filters["started_by"].add_option(
                user.username, user.get_full_name(), query=Q(created_by=user)
            )

    @override
    def get_paginator(
        self,
        per_page: int,
        *,
        orphans: int = 0,
        allow_empty_first_page: bool = True,
    ) -> WorkflowPaginator:
        return WorkflowPaginator(
            self,
            per_page=per_page,
            orphans=orphans,
            allow_empty_first_page=allow_empty_first_page,
        )


class CollectionItemTable(Table[CollectionItem]):
    """Table set up for CollectionItem."""

    category = StringColumn("Category", ordering="category")
    name = StringColumn("Name", ordering="name")
    created_at = DateTimeColumn("Created at", ordering="created_at")
    details = Column("Details")

    template_name = "web/_collection_item-list.html"
    default_order = "category"


class ArtifactTable(Table[Artifact]):
    """Table set up for Artifact."""

    category = StringColumn(
        "Type",
        ordering=Ordering(asc=["category", "id"], desc=["-category", "id"]),
    )
    name = StringColumn("Name")
    links = Column("")

    template_name = "web/_artifact-list.html"
    default_order = "category"



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Wait task to block until a user provides a signature for an upload."""

from typing import override

from debusine.client.models import LookupChildType
from debusine.server.tasks.wait import BaseWaitTask
from debusine.server.tasks.wait.models import (
    ExternalDebsignData,
    ExternalDebsignDynamicData,
)


class ExternalDebsign(
    BaseWaitTask[ExternalDebsignData, ExternalDebsignDynamicData]
):
    """Task that requests and waits for an external signature for an upload."""

    TASK_VERSION = 1

    @override
    def compute_dynamic_data(self) -> ExternalDebsignDynamicData:
        unsigned = self.work_request.lookup_single(
            self.data.unsigned, expect_type=LookupChildType.ARTIFACT
        ).artifact
        return ExternalDebsignDynamicData(unsigned_id=unsigned.id)

    @override
    def get_label(self) -> str:
        return f"wait for external debsign for {self.data.unsigned}"



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Base support for tasks that run on signing workers."""

from abc import ABCMeta
from typing import override

from django.db import close_old_connections, transaction

from debusine.artifacts.models import TaskTypes, WorkRequestResults
from debusine.tasks import BaseExternalTask
from debusine.tasks.models import BaseDynamicTaskData, BaseTaskData


class BaseSigningTask[
    TD: BaseTaskData,
    DTD: BaseDynamicTaskData = BaseDynamicTaskData,
](BaseExternalTask[TD, DTD], metaclass=ABCMeta):
    """Base class for tasks that run on signing workers."""

    TASK_TYPE = TaskTypes.SIGNING

    @override
    def execute(self) -> WorkRequestResults:
        """
        Execute task in a transaction.

        Close database connections before and after running the task to
        avoid problems with long-running worker threads.
        """
        close_old_connections()
        try:
            with transaction.atomic():
                return super().execute()
        finally:
            close_old_connections()



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.
"""Tests for the base views."""

from typing import cast
from unittest import mock

from django.core.exceptions import PermissionDenied
from django.http import Http404, HttpResponse
from django.template.response import TemplateResponse
from django.test import RequestFactory
from django.views.generic import TemplateView

from debusine.artifacts.models import CollectionCategory
from debusine.db.context import context
from debusine.db.models import Scope
from debusine.db.permissioncontext import PermissionContext
from debusine.db.playground import scenarios
from debusine.test.django import (
    AllowAll,
    DenyAll,
    TestCase,
    TestResponseType,
    override_permission,
)
from debusine.web.views.base import BaseUIView
from debusine.web.views.places import IndexPlace
from debusine.web.views.tests.utils import ViewTestMixin


class TemplateBaseUIView(BaseUIView, TemplateView):
    """BaseUIView and TemplateView to test base template rendering."""

    template_name = "web/_base_scoped.html"


class TestBaseUI(ViewTestMixin, TestCase):
    """Test the BaseUI view."""

    scenario = scenarios.DefaultContext()

    def render_base(self) -> TestResponseType:
        """Render the base template."""
        factory = RequestFactory()
        request = factory.get("/")
        assert context.user is not None
        request.user = context.user
        request.session = self.client.session
        view = TemplateBaseUIView.as_view()
        response = cast(TemplateResponse, view(request))
        response.render()
        # Django is somewhat inconsistent with naming: alias context_data to
        # context to maintain the API of test responses
        response.context = response.context_data  # type: ignore[assignment]
        return cast(TestResponseType, response)

    def test_enforce(self) -> None:
        """Test the enforce method."""
        scope = self.playground.get_default_scope()
        user = self.playground.get_default_user()
        context.set_permission_context_scoped(
            scope, PermissionContext.for_ui_user(user)
        )
        view = BaseUIView()
        with override_permission(Scope, "can_display", AllowAll):
            view.enforce(scope.can_display)
        with (
            override_permission(Scope, "can_display", DenyAll),
            self.assertRaisesRegex(
                PermissionDenied, r"playground cannot display scope debusine"
            ),
        ):
            view.enforce(scope.can_display)

    def test_set_current_workspace_wrong_name(self) -> None:
        """Test setting the current workspace to an invalid name."""
        context.set_permission_context_scoped(
            self.scenario.scope, self.scenario.context_ui()
        )
        view = BaseUIView()
        with self.assertRaises(Http404):
            view.set_current_workspace("does-not-exist")

    def _call_init_view(self, method: str, add_method: bool) -> bool:
        """Check if init_view is called for the given method."""
        factory = RequestFactory()
        request = getattr(factory, method)("/")
        request.session = self.client.session

        if add_method:
            view_cls = type(
                "TestView",
                (BaseUIView,),
                {method: lambda *args: HttpResponse("ok")},
            )
            view = view_cls.as_view()  # type: ignore[attr-defined]
        else:
            view = BaseUIView.as_view()

        with mock.patch(
            "debusine.web.views.base.BaseUIView.init_view"
        ) as init_view:
            view(request)

        return init_view.called

    def test_init_view(self) -> None:
        """Test init_view handling."""
        for method in ("get", "post", "put", "patch", "delete", "head"):
            with self.subTest(method):
                self.assertFalse(self._call_init_view(method, False))
                self.assertTrue(self._call_init_view(method, True))

        # options is always defined in views, but init_view is not called for
        # it
        self.assertFalse(self._call_init_view("options", False))

        # init_view is not called for trace
        self.assertFalse(self._call_init_view("trace", False))
        self.assertFalse(self._call_init_view("trace", True))

    def test_worker_token_denied(self) -> None:
        """Worker tokens cannot be used with web views."""
        token = self.playground.create_worker_token()

        context.set_permission_context_scoped(
            self.scenario.scope,
            PermissionContext.for_worker(token),
        )

        request = RequestFactory().get("/")
        request.session = self.client.session

        with self.assertRaisesRegex(
            PermissionDenied,
            "special purpose tokens cannot be used with web views",
        ):
            BaseUIView.as_view()(request)

    def test_work_request_token_denied(self) -> None:
        """Work request tokens cannot be used with web views."""
        token = self.playground.create_work_request_token()

        context.set_permission_context_scoped(
            self.scenario.scope,
            PermissionContext.for_work_request(token=token),
        )

        request = RequestFactory().get("/")
        request.session = self.client.session

        with self.assertRaisesRegex(
            PermissionDenied,
            "work request tokens cannot be used with web views",
        ):
            BaseUIView.as_view()(request)

    def test_get_context_data_no_workspace(self) -> None:
        """Test get_context_data with no current workspace."""
        context.set_permission_context_scoped(
            self.scenario.scope, self.scenario.context_ui()
        )
        view = BaseUIView()
        view.request = self.make_request("/")
        ctx = view.get_context_data()
        self.assertEqual(ctx["base_template"], "web/_base_scoped.html")
        place = ctx["place"]
        self.assertIsInstance(place, IndexPlace)
        self.assertEqual(place.url, "/")
        self.assertNotIn("workspace_collections", ctx)

    def test_get_context_data_with_workspace(self) -> None:
        """Test get_context_data with a current workspace."""
        self.scenario.set_current()

        self.playground.create_collection(
            name="testing", category=CollectionCategory.WORKFLOW_INTERNAL
        )

        self.playground.create_collection(
            name="bookworm", category=CollectionCategory.SUITE
        )

        view = BaseUIView()
        view.request = self.make_request("/")
        ctx = view.get_context_data()
        self.assertEqual(ctx["base_template"], "web/_base_scoped.html")
        place = ctx["place"]
        self.assertIsInstance(place, IndexPlace)
        self.assertEqual(place.url, "/")
        self.assertQuerySetEqual(
            ctx["workspace_collections"],
            list(
                self.scenario.workspace.collections.exclude(
                    category=CollectionCategory.WORKFLOW_INTERNAL
                ).order_by("name", "category")
            ),
        )

    def test_get_context_data_with_version(self) -> None:
        """get_context_data returns debusine's version if known."""
        self.scenario.set_current()
        view = BaseUIView()
        view.request = self.make_request("/")

        with mock.patch(
            "debusine.web.views.base.debusine_version", return_value="0.1.0"
        ):
            ctx = view.get_context_data()

        self.assertEqual(ctx["debusine_version"], "0.1.0")

    def test_get_context_data_without_version(self) -> None:
        """get_context_data tolerates not knowing debusine's version."""
        self.scenario.set_current()
        view = BaseUIView()
        view.request = self.make_request("/")

        with mock.patch(
            "debusine.web.views.base.debusine_version", return_value=None
        ):
            ctx = view.get_context_data()

        self.assertIsNone(ctx["debusine_version"])

    def test_navbar_not_logged_in(self) -> None:
        """Test the navbar with no user and no current workspace."""
        context.set_permission_context_scoped(
            self.scenario.scope, PermissionContext.for_ui_anonymous()
        )
        response = self.render_base()
        tree = self.assertResponseHTML(response)

        self.assertNavCommonElements(tree)
        self.assertNavNoUser(tree)
        self.assertNavNoWorkspaces(tree)
        self.assertNavNoWorkflows(tree)
        self.assertNavNoCollections(tree)
        self.assertNavNoPlumbing(tree)

    def test_navbar_no_workspace(self) -> None:
        """Test the navbar with no user and no current workspace."""
        context.set_permission_context_scoped(
            self.scenario.scope, self.scenario.context_ui()
        )

        response = self.render_base()
        tree = self.assertResponseHTML(response)

        self.assertNavCommonElements(tree)
        self.assertNavHasUser(tree, self.scenario.user)
        self.assertNavNoWorkspaces(tree)
        self.assertNavNoWorkflows(tree)
        self.assertNavNoCollections(tree)
        self.assertNavNoPlumbing(tree)

    def test_navbar_with_workspace(self) -> None:
        """Test the navbar with user and current workspace."""
        self.scenario.set_current()

        response = self.render_base()
        tree = self.assertResponseHTML(response)

        self.assertNavCommonElements(tree)
        self.assertNavHasCollections(tree, self.scenario.workspace)
        self.assertNavHasWorkflows(tree, self.scenario.workspace, [])
        self.assertNavHasWorkspaces(tree, self.scenario.workspace, [])
        self.assertNavHasUser(tree, self.scenario.user)
        self.assertNavHasPlumbing(tree)

    def test_navbar_with_workflow_templates(self) -> None:
        """Test the navbar with user and current workspace."""
        self.scenario.set_current()

        wt_foo = self.playground.create_workflow_template("foo", "noop")
        wt_bar = self.playground.create_workflow_template("bar", "noop")
        wt_baz = self.playground.create_workflow_template("baz", "noop")

        response = self.render_base()
        tree = self.assertResponseHTML(response)

        self.assertNavCommonElements(tree)
        self.assertNavHasCollections(tree, self.scenario.workspace)
        self.assertNavHasWorkflows(
            tree,
            self.scenario.workspace,
            [wt_bar, wt_baz, wt_foo],
        )
        self.assertNavHasWorkspaces(tree, self.scenario.workspace, [])
        self.assertNavHasUser(tree, self.scenario.user)

    def test_navbar_with_other_workspace(self) -> None:
        """Test the navbar with current plus other workspaces to list."""
        self.scenario.set_current()
        b = self.playground.create_workspace(name="b", public=True)
        a = self.playground.create_workspace(name="a", public=True)
        self.playground.create_workspace(name="c")

        response = self.render_base()
        tree = self.assertResponseHTML(response)
        self.assertEqual(response.context["other_workspaces"], [a, b])

        self.assertNavCommonElements(tree)
        self.assertNavHasCollections(tree, self.scenario.workspace)
        self.assertNavHasWorkspaces(tree, self.scenario.workspace, [a, b])
        self.assertNavHasUser(tree, self.scenario.user)

    def test_footer_with_version(self) -> None:
        """The footer shows debusine's version if known."""
        self.scenario.set_current()
        with mock.patch(
            "debusine.web.views.base.debusine_version", return_value="0.1.0"
        ):
            response = self.render_base()
        tree = self.assertResponseHTML(response)

        self.assertTextContentEqual(
            tree.xpath("//footer/div[@id='footer-software']")[0],
            "Debusine software project: Documentation • Bugs • Code • "
            "Contributing • 0.1.0",
        )

    def test_footer_without_version(self) -> None:
        """The footer skips debusine's version if not known."""
        self.scenario.set_current()
        with mock.patch(
            "debusine.web.views.base.debusine_version", return_value=None
        ):
            response = self.render_base()
        tree = self.assertResponseHTML(response)

        self.assertTextContentEqual(
            tree.xpath("//footer/div[@id='footer-software']")[0],
            "Debusine software project: Documentation • Bugs • Code • "
            "Contributing",
        )

    def test_footer_custom_replace(self) -> None:
        """Footer replaced by templates installed by site admins."""
        self.scenario.set_current()
        tree = self.assertResponseHTML(self.render_base())
        footer = self.assertHasElement(
            tree, "//footer/*/div[@id='footer-status']"
        )
        self.assertTextContentEqual(
            footer, "Status: Worker pools • Workers • Task status"
        )

        with self.custom_template("web/footer.html") as template:
            template.write_text("CUSTOM")
            tree = self.assertResponseHTML(self.render_base())
            footer = self.assertHasElement(tree, "//footer")
            self.assertTextContentEqual(footer, "CUSTOM")

    def test_footer_custom_extend(self) -> None:
        """Footer extended by templates installed by site admins."""
        self.scenario.set_current()
        tree = self.assertResponseHTML(self.render_base())
        footer = self.assertHasElement(
            tree, "//footer/*/div[@id='footer-status']"
        )
        self.assertTextContentEqual(
            footer, "Status: Worker pools • Workers • Task status"
        )

        with self.custom_template("web/footer.html") as template:
            template.write_text(
                "{% extends 'web/footer.html' %}"
                "{% block footer_status %}{{block.super}}"
                "<div id='footer-testing'>Testing</div>{% endblock %}"
            )
            tree = self.assertResponseHTML(self.render_base())
            footer = self.assertHasElement(
                tree, "//footer/div[@id='footer-utils']"
            )
            self.assertTextContentEqual(
                footer,
                "Status: Worker pools • Workers • Task status"
                " Testing"
                " Theme: dark • light • auto",
            )

    def test_theme_selection(self) -> None:
        self.scenario.set_current()
        view = BaseUIView()
        view.request = self.make_request("/")
        ctx = view.get_context_data()
        self.assertIsNone(ctx["theme"])
        view.request.COOKIES["theme"] = "foo"
        ctx = view.get_context_data()
        self.assertEqual(ctx["theme"], "foo")



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Views for the server application's system health status."""

from rest_framework import status
from rest_framework.request import Request
from rest_framework.response import Response

from debusine.db.models.permissions import anonymous_access
from debusine.server.status import WorkerStatus
from debusine.server.views.base import BaseAPIView


class ServiceStatusView(BaseAPIView):
    """View used to get the service status."""

    def get(self, request: Request) -> Response:  # noqa: ARG002, U100
        """Return system health status."""
        self.enforce(anonymous_access)

        health_status = WorkerStatus.get_status()
        data = health_status.model_dump(mode="json")
        return Response(data, status=status.HTTP_200_OK)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Common code for tabular output tests."""

import contextlib
import unittest
from collections.abc import Generator
from typing import Any
from unittest import mock

from rich.table import Table


class TableOutput:
    """Holder for assertPrintTable result."""

    def __init__(self, table: Table | None = None) -> None:
        """Make space to hold a table."""
        self.table: Table | None = table

    def col(self, index: int) -> list[Any]:
        """Return a list of values for a table column."""
        assert self.table is not None
        return list(self.table.columns[index].cells)


class TabularOutputTests(unittest.TestCase):
    """Common functions for testing tabular output."""

    @contextlib.contextmanager
    def assertPrintsTable(self) -> Generator[TableOutput]:
        """Check that a rich.Table is printed."""
        output = TableOutput()
        with mock.patch("rich.print") as rprint:
            yield output
        rprint.assert_called_once()
        output.table = rprint.call_args.args[0]
        self.assertIsInstance(output.table, Table)

    @contextlib.contextmanager
    def assertPrintsTables(self) -> Generator[list[TableOutput]]:
        """Check that multiple rich.Tables are printed."""
        output: list[TableOutput] = []
        with mock.patch("rich.print") as rprint:
            yield output
        rprint.assert_called()
        for call in rprint.call_args_list:
            table = call.args[0]
            self.assertIsInstance(table, Table)
            output.append(TableOutput(table=table))



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.
"""Debusine server controller."""

import hashlib
import logging

import requests
import yaml
from prometheus_client.parser import text_string_to_metric_families

from utils import common
from utils.common import Configuration, RunResult
from utils.waiter import Waiter

logger = logging.getLogger(__name__)


class DebusineServer:
    """Interacts with the debusine-server: execute debusine-admin commands."""

    @staticmethod
    def wait_for_server_ready() -> bool:
        """
        Wait up to 30 seconds for the server to be ready.

        Return True/False depending on if the server is ready or not.
        """

        def is_ready() -> bool:
            """Return True if {self._api_url}/ is available."""
            response = requests.get(f"{Configuration.get_base_url()}/")
            return response.status_code < 400

        return Waiter.wait_for_success(30, is_ready)

    @staticmethod
    def restart() -> RunResult:
        """Restart (via systemctl) debusine-server."""
        return common.run(["systemctl", "restart", "debusine-server"])

    @staticmethod
    def execute_command(
        command: str,
        *args: str,
        user: str = Configuration.DEBUSINE_SERVER_USER,
        stdin: str | None = None,
    ) -> RunResult:
        """Execute a debusine server management command."""
        cmd = ["debusine-admin", command, *args]
        result = common.run_as(user, cmd, stdin=stdin)

        if result.stdout:
            logger.info("stdout: %s", result.stdout)
        if result.stderr:
            logger.info("stderr: %s", result.stderr)

        return result

    @classmethod
    def verify_worker(
        cls,
        token: str,
        connected: bool,
        enabled: bool,
        list_workers: str | None = None,
    ) -> bool:
        """
        Return True if ``worker list`` has token with the specified state.

        :param token: token that is being verified
        :param connected: True if it is expected that the worker is connected
        :param enabled: True if it is expected that the worker's token enabled
        :param list_workers: output of the command ``worker list --yaml`` or
          None.  If None it executes ``worker list --yaml``
        :return: True if a token is connected and enabled as per connected
          and enabled parameters
        """
        if list_workers is None:
            list_workers_output = cls.execute_command(
                "worker", "list", "--yaml"
            ).stdout

        token_hash = hashlib.sha256(token.encode()).hexdigest()
        for worker in yaml.safe_load(list_workers_output):
            if worker["hash"] == token_hash:
                return (
                    bool(worker["connected"]) == connected
                    and worker["enabled"] == enabled
                )

        return False

    @classmethod
    def wait_for_worker_connected(cls, token: str) -> bool:
        """Wait up to 30 seconds for the worker to be connected."""

        def token_is_connected() -> bool:
            return cls.verify_worker(token, connected=True, enabled=True)

        return Waiter.wait_for_success(30, token_is_connected)

    @classmethod
    def get_metric(
        cls, *, metric_name: str, sample_name: str, **labels: str
    ) -> float | None:
        """Get the value of a metric from the server."""
        response = requests.get(
            f"{Configuration.get_base_url()}/api/1.0/open-metrics/"
        )
        response.raise_for_status()
        for metric in text_string_to_metric_families(response.text):
            if metric.name == metric_name:
                for sample in metric.samples:
                    if sample.name == sample_name and sample.labels == labels:
                        return sample.value
        return None



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the scopes models."""

from collections.abc import Collection as AbcCollection
from typing import Any, ClassVar, cast, override

from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.db import IntegrityError, transaction
from django.test import RequestFactory, override_settings

from debusine.db.context import context
from debusine.db.models import FileInStore, FileStore, Group, Scope
from debusine.db.models.auth import system_user
from debusine.db.models.permissions import Allow
from debusine.db.models.scopes import (
    ScopeRole,
    ScopeRoleBase,
    ScopeRoles,
    is_valid_scope_name,
)
from debusine.db.models.tests.utils import RolesTestCase
from debusine.db.permissioncontext import PermissionContext
from debusine.db.playground import scenarios
from debusine.test.django import TestCase


class ScopeRolesTests(RolesTestCase[ScopeRoles]):
    """Tests for the ScopeRoles class."""

    roles_class = ScopeRoles

    def test_invariants(self) -> None:
        self.assertRolesInvariants()

    def test_from_iterable(self) -> None:
        for value, expected in (
            ([], []),
            (["owner"], [Scope.Roles.OWNER]),
            ([Scope.Roles.OWNER], [Scope.Roles.OWNER]),
            (["owner", "owner", Scope.Roles.OWNER], [Scope.Roles.OWNER]),
            (["observer"], [Scope.Roles.OBSERVER]),
            (["observer", Scope.Roles.OWNER], [Scope.Roles.OWNER]),
        ):
            with self.subTest(value=repr(value)):
                self.assertEqual(
                    Scope.Roles.from_iterable(value),
                    frozenset(expected),
                )

    def test_choices(self) -> None:
        self.assertChoices()

    def test_values(self) -> None:
        self.assertEqual(ScopeRoles.OWNER.label, "Owner")
        self.assertEqual(
            ScopeRoles.OWNER.implied_by_scope_roles,
            frozenset([ScopeRoles.OWNER]),
        )
        self.assertEqual(
            ScopeRoles.OWNER.description,
            "Full administrative access",
        )

        self.assertEqual(ScopeRoles.OBSERVER.label, "Observer")
        self.assertEqual(
            ScopeRoles.OBSERVER.implied_by_scope_roles,
            frozenset([ScopeRoles.OWNER, ScopeRoles.OBSERVER]),
        )
        self.assertEqual(
            ScopeRoles.OBSERVER.description,
            "Read-only access to all workspaces",
        )

    def test_init_invalid_implication(self) -> None:
        r = ScopeRoleBase("test", implied_by=[Scope.Roles.OWNER])
        with self.assertRaisesRegex(
            ImproperlyConfigured,
            r"Scope roles do not support implications"
            r" by <enum 'ScopeRoles'>",
        ):
            r._setup()


class ScopeManagerTests(TestCase):
    """Tests for the ScopeManager class."""

    scenario = scenarios.DefaultScopeUser()

    def test_get_roles_model(self) -> None:
        """Test the get_roles_model method."""
        self.assertIs(Scope.objects.get_roles_model(), ScopeRole)


@override_settings(LANGUAGE_CODE="en-us")
class ScopeTests(TestCase):
    """Tests for the Scope class."""

    scenario = scenarios.DefaultScopeUser()

    def assertHasRole(
        self,
        *,
        pc: PermissionContext | None = None,
        scope: Scope | None = None,
        hits_database: bool = False,
    ) -> None:
        """Ensure the user has OWNER role on the scope."""
        scope = scope or self.scenario.scope
        pc = pc or self.scenario.context_ui()
        expected_queries = 1 if hits_database else 0
        with self.assertNumQueries(expected_queries):
            self.assertTrue(scope.has_role(pc, Scope.Roles.OWNER))

    def assertHasNoRole(
        self,
        *,
        pc: PermissionContext | None = None,
        scope: Scope | None = None,
        hits_database: bool = False,
    ) -> None:
        """Ensure the user does not have OWNER role on the scope."""
        scope = scope or self.scenario.scope
        pc = pc or self.scenario.context_ui()
        expected_queries = 1 if hits_database else 0
        with self.assertNumQueries(expected_queries):
            self.assertFalse(scope.has_role(pc, Scope.Roles.OWNER))

    def test_get_ui(self) -> None:
        from debusine.web.views.ui.scopes import ScopeUI

        request = RequestFactory().get("/")
        ui = self.scenario.scope.ui(request)
        self.assertIsInstance(ui, ScopeUI)

    def test_create(self) -> None:
        """Test basic behavior."""
        scope = Scope.objects.create(name="test")
        self.assertEqual(scope.name, "test")
        self.assertEqual(str(scope), "test")

    def test_unique(self) -> None:
        """Check that scope names and labels are unique."""
        Scope.objects.create(name="test", label="Test")
        with (
            transaction.atomic(),
            self.assertRaisesRegex(
                IntegrityError,
                # Only match constraint name to support non-english locales
                # "duplicate key value violates unique constraint"
                r"(?s)duplicate key value violates unique constraint.+"
                r"Key \(name\)=\(test\) already exists",
            ),
        ):
            Scope.objects.create(name="test", label="Other test")
        with (
            transaction.atomic(),
            self.assertRaisesRegex(
                IntegrityError,
                # Only match constraint name to support non-english locales
                # "duplicate key value violates unique constraint"
                r"(?s)duplicate key value violates unique constraint.+"
                r"Key \(label\)=\(Test\) already exists",
            ),
        ):
            Scope.objects.create(name="test1", label="Test")
        Scope.objects.create(name="test1", label="Test1")

    def test_fallback(self) -> None:
        """The fallback scope for migrations exists."""
        scope = Scope.objects.get(name="debusine")
        self.assertEqual(scope.name, "debusine")

    def test_is_valid_scope_name(self) -> None:
        """Test is_valid_scope_name."""
        valid_names = (
            "debian",
            "debian-lts",
            "debusine",
            "c",
            "c++",
            "foo_bar",
            "foo.bar",
            "foo+bar",
            "tail_",
            "c--",
        )
        invalid_names = (
            "api",
            "admin",
            "user",
            "+tag",
            "-tag",
            ".profile",
            "_reserved",
            "foo:bar",
        )

        for name in valid_names:
            with self.subTest(name=name):
                self.assertTrue(is_valid_scope_name(name))

        for name in invalid_names:
            with self.subTest(name=name):
                self.assertFalse(is_valid_scope_name(name))

    def test_scope_name_validation(self) -> None:
        """Test validation for scope names."""
        Scope(name="foo", label="Foo").full_clean()
        Scope(name="foo_", label="Foo").full_clean()
        with self.assertRaises(ValidationError) as exc:
            Scope(name="_foo", label="Foo").full_clean()
        self.assertEqual(
            exc.exception.message_dict,
            {'name': ["'_foo' is not a valid scope name"]},
        )

    def test_get_group_roles_not_authenticated(self) -> None:
        """Test Scope.get_group_roles when not authenticated."""
        scope = Scope.objects.create(name="Scope")
        self.assertQuerySetEqual(
            scope.get_group_roles(PermissionContext.for_ui_anonymous()), []
        )

    def test_get_group_roles(self) -> None:
        """Test Scope.get_group_roles."""
        scope = Scope.objects.create(name="Scope")
        group = self.playground.create_group_role(scope, ScopeRole.Roles.OWNER)

        user = self.scenario.user
        pc = PermissionContext.for_ui_user(user)
        self.assertQuerySetEqual(scope.get_group_roles(pc), [])

        self.playground.add_user(group, user)
        self.assertEqual(
            list(scope.get_group_roles(pc)), [ScopeRole.Roles.OWNER]
        )

    def test_get_group_roles_multiple_groups(self) -> None:
        """Test Scope.get_group_roles with the user in multiple groups."""
        scope = Scope.objects.create(name="Scope")
        group1 = self.playground.create_group_role(
            scope, ScopeRole.Roles.OWNER, name="Group1"
        )
        group2 = self.playground.create_group_role(
            scope, ScopeRole.Roles.OWNER, name="Group2"
        )

        user = self.scenario.user
        pc = PermissionContext.for_ui_user(user)
        self.assertQuerySetEqual(scope.get_group_roles(pc), [])

        self.playground.add_user(group1, user)
        self.assertEqual(
            list(scope.get_group_roles(pc)), [ScopeRole.Roles.OWNER]
        )

        self.playground.add_user(group2, user)
        self.assertEqual(
            list(scope.get_group_roles(pc)), [ScopeRole.Roles.OWNER]
        )

    def test_get_group_roles_extra_groups(self) -> None:
        pc = self.playground.create_pc_with_extra_groups(
            self.scenario.scope_owners
        )
        self.assertEqual(
            list(self.scenario.scope.get_group_roles(pc)),
            [ScopeRole.Roles.OWNER],
        )

    def test_with_role(self) -> None:
        self.scenario.scope_owners.users.add(self.scenario.user)
        other = self.playground.create_user("other")
        for pc, expected in (
            (PermissionContext.for_ui_anonymous(), []),
            (PermissionContext.for_ui_user(other), []),
            (self.scenario.context_ui(), [self.scenario.scope]),
            (
                self.playground.create_pc_with_extra_groups(
                    self.scenario.scope_owners, user=other
                ),
                [self.scenario.scope],
            ),
        ):
            with self.subTest(pc=repr(pc)):
                self.assertQuerySetEqual(
                    Scope.objects.with_role(pc, Scope.Roles.OWNER),
                    expected,
                )

    def test_has_role_anonymous_user(self) -> None:
        """Text has_role with anonymous users."""
        with self.assertNumQueries(0):
            self.assertFalse(
                self.scenario.scope.has_role(
                    PermissionContext.for_ui_anonymous(), Scope.Roles.OWNER
                )
            )

    def test_has_role_system_user(self) -> None:
        """The SYSTEM_USER has implicit scope ownership."""
        pc = PermissionContext.for_ui_user(system_user())
        self.assertTrue(self.scenario.scope.has_role(pc, Scope.Roles.OWNER))
        self.assertQuerySetEqual(
            Scope.objects.with_role(pc, Scope.Roles.OWNER),
            cast(AbcCollection[Any], Scope.objects.all()),
        )

    def test_has_role_no_context(self) -> None:
        """Test has_role with no context set."""
        self.assertHasNoRole(hits_database=True)
        self.scenario.scope_owners.users.add(self.scenario.user)
        self.assertHasRole(hits_database=True)

    def test_has_role_user_not_in_context(self) -> None:
        """Test has_role with a different user than in context."""
        self.scenario.set_current()
        test = self.playground.create_user("test")
        pc = PermissionContext.for_ui_user(test)
        self.assertHasNoRole(pc=pc, hits_database=True)
        self.scenario.scope_owners.users.add(test)
        self.assertHasRole(pc=pc, hits_database=True)

    def test_has_role_scope_not_in_context(self) -> None:
        """Test has_role with a different scope than in context."""
        self.scenario.set_current()
        scope = self.playground.get_or_create_scope("test")
        self.assertHasNoRole(scope=scope, hits_database=True)
        self.playground.create_group_role(
            scope, Scope.Roles.OWNER, users=[self.scenario.user]
        )
        self.assertHasRole(scope=scope, hits_database=True)

    def test_has_role_from_context(self) -> None:
        """Test has_role with information from context."""
        with context.local():
            self.scenario.set_current()
            self.assertHasNoRole(hits_database=False)

        self.scenario.scope_owners.users.add(self.scenario.user)

        with context.local():
            self.scenario.set_current()
            self.assertHasRole(hits_database=False)

    def test_has_role_extra_groups_no_context(self) -> None:
        pc = self.playground.create_pc_with_extra_groups(
            self.scenario.scope_owners
        )
        self.assertHasRole(pc=pc, hits_database=True)

    def test_has_role_extra_groups_user_not_in_context(self) -> None:
        user = self.playground.create_user("test")
        pc = self.playground.create_pc_with_extra_groups(
            self.scenario.scope_owners, user=user
        )
        self.scenario.set_current()
        self.assertHasRole(pc=pc, hits_database=True)

    def test_has_role_extra_groups_scope_not_in_context(self) -> None:
        pc = self.playground.create_pc_with_extra_groups(
            self.scenario.scope_owners
        )
        context.set_permission_context_scoped(scope=self.scenario.scope, pc=pc)
        scope = self.playground.get_or_create_scope("test")
        self.assertHasNoRole(scope=scope, hits_database=True)

    def test_has_role_extra_groups_from_context(self) -> None:
        """Test has_role with information from context."""
        pc = self.playground.create_pc_with_extra_groups(
            self.scenario.scope_owners
        )
        context.set_permission_context_scoped(scope=self.scenario.scope, pc=pc)
        self.assertHasRole(pc=pc, hits_database=False)

    def test_upload_file_stores(self) -> None:
        """`upload_file_stores` returns stores in the correct order."""
        fileobj = self.playground.create_file(contents=b"test")
        scope = self.scenario.scope
        first_store = scope.file_stores.get()
        read_only_store = FileStore.objects.create(
            name="read_only",
            backend=FileStore.BackendChoices.MEMORY,
            configuration={"name": "read_only"},
        )
        scope.file_stores.add(
            read_only_store, through_defaults={"read_only": True}
        )
        other_store = FileStore.objects.create(
            name="other",
            backend=FileStore.BackendChoices.MEMORY,
            configuration={"name": "other"},
        )
        max_size_store = FileStore.objects.create(
            name="full",
            backend=FileStore.BackendChoices.MEMORY,
            configuration={"name": "full"},
            max_size=5,
        )
        scope.file_stores.add(other_store, max_size_store)

        self.assertQuerySetEqual(
            scope.upload_file_stores(fileobj),
            [first_store, other_store, max_size_store],
        )
        self.assertEqual(
            scope.upload_file_backend(fileobj).db_store,
            first_store.get_backend_object().db_store,
        )

        # If a store's max_size would be exceeded by adding the new file, it
        # is not eligible to accept the upload.
        self.playground.create_file_in_backend(
            max_size_store.get_backend_object(), contents=b"abcd"
        )

        self.assertEqual(
            (
                FileStore.objects.annotate_current_total_size()
                .values_list("current_total_size", flat=True)
                .get(pk=max_size_store.pk)
            ),
            4,
        )
        self.assertQuerySetEqual(
            scope.upload_file_stores(fileobj), [first_store, other_store]
        )
        self.assertEqual(
            scope.upload_file_backend(fileobj).db_store,
            first_store.get_backend_object().db_store,
        )

        # The default store has an upload priority of 100.  Test ordering
        # relative to that.
        other_store_extra = scope.filestoreinscope_set.get(
            file_store=other_store
        )
        other_store_extra.upload_priority = 200
        other_store_extra.save()

        self.assertQuerySetEqual(
            scope.upload_file_stores(fileobj), [other_store, first_store]
        )
        self.assertEqual(
            scope.upload_file_backend(fileobj).db_store,
            other_store.get_backend_object().db_store,
        )

        other_store_extra.upload_priority = 90
        other_store_extra.save()

        self.assertQuerySetEqual(
            scope.upload_file_stores(fileobj), [first_store, other_store]
        )
        self.assertEqual(
            scope.upload_file_backend(fileobj).db_store,
            first_store.get_backend_object().db_store,
        )

    def test_upload_file_stores_enforce_soft_limits(self) -> None:
        """`upload_file_stores` can optionally enforce soft limits."""
        fileobj = self.playground.create_file(contents=b"test")
        scope = self.scenario.scope
        first_store = scope.file_stores.get()
        soft_max_size_store = FileStore.objects.create(
            name="near-full",
            backend=FileStore.BackendChoices.MEMORY,
            configuration={"name": "near-full"},
            soft_max_size=5,
            max_size=10,
        )
        scope.file_stores.add(soft_max_size_store)

        # If a store's soft_max_size would be exceeded by adding the new
        # file, and soft limits are enforced, it is not eligible to accept
        # the upload.
        self.playground.create_file_in_backend(
            soft_max_size_store.get_backend_object(), contents=b"abcd"
        )

        self.assertEqual(
            (
                FileStore.objects.annotate_current_total_size()
                .values_list("current_total_size", flat=True)
                .get(pk=soft_max_size_store.pk)
            ),
            4,
        )
        self.assertQuerySetEqual(
            scope.upload_file_stores(fileobj),
            [first_store, soft_max_size_store],
        )
        self.assertQuerySetEqual(
            scope.upload_file_stores(fileobj, enforce_soft_limits=True),
            [first_store],
        )
        self.assertEqual(
            scope.upload_file_backend(fileobj).db_store,
            first_store.get_backend_object().db_store,
        )

    def test_upload_file_stores_include_write_only(self) -> None:
        """`upload_file_stores` can optionally include write-only stores."""
        fileobj = self.playground.create_file(contents=b"test")
        scope = self.scenario.scope
        first_store = scope.file_stores.get()
        write_only_store = FileStore.objects.create(
            name="write_only",
            backend=FileStore.BackendChoices.MEMORY,
            configuration={"name": "write_only"},
        )
        scope.file_stores.add(
            write_only_store, through_defaults={"write_only": True}
        )

        self.assertQuerySetEqual(
            scope.upload_file_stores(fileobj), [first_store]
        )
        self.assertQuerySetEqual(
            scope.upload_file_stores(fileobj, include_write_only=True),
            [first_store, write_only_store],
        )

    def test_download_file_stores(self) -> None:
        """`download_file_stores` returns stores in the correct order."""
        fileobj = self.playground.create_file()
        scope = self.scenario.scope
        first_store = scope.file_stores.get()
        write_only_store = FileStore.objects.create(
            name="write_only",
            backend=FileStore.BackendChoices.MEMORY,
            configuration={"name": "write_only"},
        )
        scope.file_stores.add(
            write_only_store, through_defaults={"write_only": True}
        )
        other_store = FileStore.objects.create(
            name="other",
            backend=FileStore.BackendChoices.MEMORY,
            configuration={"name": "other"},
        )
        scope.file_stores.add(other_store)

        self.assertQuerySetEqual(scope.download_file_stores(fileobj), [])
        self.assertRaises(IndexError, scope.download_file_backend, fileobj)

        FileInStore.objects.create(file=fileobj, store=first_store)
        FileInStore.objects.create(file=fileobj, store=write_only_store)

        self.assertQuerySetEqual(
            scope.download_file_stores(fileobj), [first_store]
        )
        self.assertEqual(
            scope.download_file_backend(fileobj).db_store,
            first_store.get_backend_object().db_store,
        )

        FileInStore.objects.create(file=fileobj, store=other_store)

        self.assertQuerySetEqual(
            scope.download_file_stores(fileobj), [first_store, other_store]
        )
        self.assertEqual(
            scope.download_file_backend(fileobj).db_store,
            first_store.get_backend_object().db_store,
        )

        # The default store has upload and download priorities of 100.  Test
        # ordering relative to that.
        other_store_extra = scope.filestoreinscope_set.get(
            file_store=other_store
        )
        other_store_extra.download_priority = 200
        other_store_extra.save()

        self.assertQuerySetEqual(
            scope.download_file_stores(fileobj), [other_store, first_store]
        )
        self.assertEqual(
            scope.download_file_backend(fileobj).db_store,
            other_store.get_backend_object().db_store,
        )

        other_store_extra.download_priority = 100
        other_store_extra.upload_priority = 200
        other_store_extra.save()

        self.assertQuerySetEqual(
            scope.download_file_stores(fileobj), [other_store, first_store]
        )
        self.assertEqual(
            scope.download_file_backend(fileobj).db_store,
            other_store.get_backend_object().db_store,
        )

        other_store_extra.upload_priority = 90
        other_store_extra.save()

        self.assertQuerySetEqual(
            scope.download_file_stores(fileobj), [first_store, other_store]
        )
        self.assertEqual(
            scope.download_file_backend(fileobj).db_store,
            first_store.get_backend_object().db_store,
        )

    def test_can_display(self) -> None:
        """Test the can_display predicate."""
        scope = self.scenario.scope
        scope2 = Scope.objects.create(name="Scope2")
        for pc in (
            PermissionContext.for_ui_anonymous(),
            self.scenario.context_ui(),
        ):
            with self.subTest(user=repr(pc.user)):
                self.assertTrue(scope.can_display(pc))
                self.assertTrue(scope2.can_display(pc))
                self.assertQuerySetEqual(
                    Scope.objects.can_display(pc),
                    [scope, scope2],
                    ordered=False,
                )

    def test_can_create_workspace(self) -> None:
        self.assertRolePredicate(
            self.scenario.scope,
            "can_create_workspace",
            Scope.Roles.OWNER,
            work_request=Allow.NEVER,
        )


class ScopeRoleTests(TestCase):
    """Tests for the ScopeRole class."""

    scope1: ClassVar[Scope]
    scope2: ClassVar[Scope]
    group1: ClassVar[Group]
    group2: ClassVar[Group]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.scope1 = cls.playground.get_or_create_scope(name="Scope1")
        cls.scope2 = cls.playground.get_or_create_scope(name="Scope2")
        cls.group1 = cls.playground.create_group("Group1", scope=cls.scope1)
        cls.group2 = cls.playground.create_group("Group2", scope=cls.scope2)

    def test_str(self) -> None:
        """Test stringification."""
        sr = ScopeRole(
            group=self.group1, resource=self.scope1, role=ScopeRole.Roles.OWNER
        )
        self.assertEqual(str(sr), "Scope1/Group1─owner⟶Scope1")

    def test_assign_multiple_groups(self) -> None:
        """Multiple group can share a role on a scope."""
        ScopeRole.objects.create(
            group=self.group1, resource=self.scope1, role=ScopeRole.Roles.OWNER
        )
        ScopeRole.objects.create(
            group=self.group2, resource=self.scope1, role=ScopeRole.Roles.OWNER
        )

    def test_assign_multiple_scopes(self) -> None:
        """A group can be given a role in different scopes."""
        ScopeRole.objects.create(
            group=self.group1, resource=self.scope1, role=ScopeRole.Roles.OWNER
        )
        ScopeRole.objects.create(
            group=self.group1, resource=self.scope2, role=ScopeRole.Roles.OWNER
        )

    def test_on_delete_cascade(self) -> None:
        """Test on delete cascade behaviour."""
        assignment = ScopeRole.objects.create(
            group=self.group1, resource=self.scope1, role=ScopeRole.Roles.OWNER
        )
        self.group1.delete()
        self.assertFalse(ScopeRole.objects.filter(pk=assignment.pk).exists())

        # We cannot test this case, because self.group2 references self.scope2
        # and prevents its deletion
        # assignment = ScopeRole.objects.create(
        #     group=self.group2, resource=self.scope2,
        #     role=ScopeRole.Roles.OWNER
        # )
        # self.scope2.delete()
        # self.assertFalse(ScopeRole.objects.filter(pk=assignment.pk).exists())


@override_settings(LANGUAGE_CODE="en-us")
class FileStoreInScopeTests(TestCase):
    """Tests for the FileStoreInScope class."""

    scenario = scenarios.DefaultScopeUser()

    def test_str(self) -> None:
        """`FileStoreInScope.__str__` returns basic information."""
        scope = self.scenario.scope
        file_store = scope.file_stores.get()
        file_store_extra = scope.filestoreinscope_set.get()

        self.assertEqual(
            file_store_extra.__str__(), f"{scope}/{file_store.name}"
        )

    def test_instance_wide(self) -> None:
        """Instance-wide file stores can be used by multiple scopes."""
        scope = self.scenario.scope
        other_scope = self.playground.get_or_create_scope(name="other")
        file_store = FileStore.objects.create(
            name="instance_wide",
            backend=FileStore.BackendChoices.MEMORY,
            configuration={"name": "instance_wide"},
        )

        scope.file_stores.add(file_store)
        other_scope.file_stores.add(file_store)

        self.assertTrue(scope.file_stores.filter(pk=file_store.pk).exists())
        self.assertTrue(
            other_scope.file_stores.filter(pk=file_store.pk).exists()
        )

    def test_non_instance_wide(self) -> None:
        """Non-instance-wide file stores can only be used by a single scope."""
        scope = self.scenario.scope
        other_scope = self.playground.get_or_create_scope(name="other")
        file_store = FileStore.objects.create(
            name="just_one_scope",
            backend=FileStore.BackendChoices.MEMORY,
            configuration={"name": "just_one_scope"},
            instance_wide=False,
        )

        scope.file_stores.add(file_store)
        with self.assertRaisesRegex(
            IntegrityError,
            r'duplicate key value violates unique constraint '
            r'"db_filestoreinscope_unique_file_store_not_instance_wide"',
        ):
            other_scope.file_stores.add(file_store)

    def test_change_instance_wide(self) -> None:
        """Changing `FileStore.instance_wide` checks constraints."""
        scope = self.scenario.scope
        other_scope = self.playground.get_or_create_scope(name="other")
        default_file_store = scope.file_stores.get()
        file_store = FileStore.objects.create(
            name="instance_wide",
            backend=FileStore.BackendChoices.MEMORY,
            configuration={"name": "instance_wide"},
        )

        scope.file_stores.add(file_store)
        file_store.instance_wide = False
        file_store.save()

        # The changed instance_wide value is mirrored to the field used by
        # the constraint.
        self.assertFalse(
            scope.filestoreinscope_set.get(
                file_store=file_store
            )._file_store_instance_wide
        )

        # The default file store's instance_wide setting is untouched.
        self.assertTrue(default_file_store.instance_wide)
        self.assertTrue(
            scope.filestoreinscope_set.get(
                file_store=default_file_store
            )._file_store_instance_wide
        )

        # Changing instance_wide fails if the constraint would be violated.
        file_store.instance_wide = True
        file_store.save()
        other_scope.file_stores.add(file_store)
        file_store.instance_wide = False
        with self.assertRaisesRegex(
            IntegrityError,
            r'duplicate key value violates unique constraint '
            r'"db_filestoreinscope_unique_file_store_not_instance_wide"',
        ):
            file_store.save()



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""debusine collection views."""

from functools import cached_property
from typing import Any, override

from django.db.models import Count, Q, QuerySet
from django.http import Http404, HttpResponse, HttpResponseRedirect
from django.urls import reverse

from debusine.db.context import context
from debusine.db.models import Collection, CollectionItem
from debusine.db.models.collections import CollectionRelation
from debusine.web.forms import (
    CollectionCreateForm,
    CollectionSearchForm,
    CollectionUpdateForm,
    ModelFormBase,
)
from debusine.web.views.base import (
    CreateViewBase,
    DeleteViewBase,
    DetailViewBase,
    FormMixinBase,
    ListViewBase,
    UpdateViewBase,
    WorkspaceView,
)
from debusine.web.views.places import Place, ResourcePlace
from debusine.web.views.table import TableMixin
from debusine.web.views.tables import CollectionItemTable
from debusine.web.views.view_utils import highlight_as_yaml


class CollectionListView(WorkspaceView, ListViewBase[Collection]):
    """List collections."""

    model = Collection
    template_name = "web/collection-list.html"
    context_object_name = "collection_list"
    ordering = ["category", "name"]

    @override
    def get_place(self) -> Place:
        return (
            context.require_workspace().ui(self.request).place_collection_list
        )

    @override
    def get_queryset(self) -> QuerySet[Collection]:
        """Filter collection by accessible workspace."""
        queryset = super().get_queryset()
        return queryset.filter(workspace=context.workspace).exclude(
            category="debusine:workflow-internal"
        )


class CollectionCategoryListView(WorkspaceView, ListViewBase[Collection]):
    """List collections with a given category."""

    model = Collection
    template_name = "web/collection-category-list.html"
    context_object_name = "collection_list"
    ordering = ["name"]

    @override
    def get_place(self) -> Place:
        workspace = context.require_workspace()
        return ResourcePlace(
            title=f"{workspace.name} {self.kwargs['ccat']} collections",
            url=reverse(
                "workspaces:collections:category_list",
                kwargs={"wname": workspace.name, "ccat": self.kwargs["ccat"]},
            ),
            breadcrumb=self.kwargs["ccat"],
        )

    @override
    def get_queryset(self) -> QuerySet[Collection]:
        """Filter collection by accessible workspace."""
        queryset = super().get_queryset()
        return (
            queryset.filter(workspace=context.workspace)
            .exclude(category="debusine:workflow-internal")
            .filter(category=self.kwargs["ccat"])
        )

    @override
    def get_context_data(self, *args: Any, **kwargs: Any) -> dict[str, Any]:
        """Return context_data with category."""
        context = super().get_context_data(**kwargs)
        context["category"] = self.kwargs["ccat"]
        return context


class CollectionCreateView(
    WorkspaceView, CreateViewBase[Collection, CollectionCreateForm]
):
    """Create a collection."""

    form_class = CollectionCreateForm
    template_name = "web/collection-create.html"

    @override
    def init_view(self) -> None:
        """Enforce permissions."""
        super().init_view()
        self.enforce(context.require_workspace().can_create_collections)

    @override
    def get_place(self) -> Place:
        return (
            context.require_workspace().ui(self.request).place_collection_create
        )


class CollectionViewMixin(WorkspaceView):
    """Common functions for collection-specific views."""

    @cached_property
    def collection(self) -> Collection:
        """Collection for this request."""
        try:
            return Collection.objects.can_display(context.pc).get(
                workspace=context.workspace,
                name=self.kwargs["cname"],
                category=self.kwargs["ccat"],
            )
        except Collection.DoesNotExist:
            raise Http404(
                f"{self.kwargs['cname']}@{self.kwargs['ccat']}"
                " collection not found"
            )

    @override
    def get_context_data(self, *args: Any, **kwargs: Any) -> dict[str, Any]:
        """Return context_data with collection."""
        context = super().get_context_data(**kwargs)
        context["collection"] = self.collection
        return context


class CollectionDetailView(CollectionViewMixin, DetailViewBase[Collection]):
    """Show a collection detail."""

    model = Collection
    template_name = "web/collection-detail.html"
    context_object_name = "collection"

    @override
    def get_place(self) -> Place:
        return self.object.ui(self.request).place

    @override
    def get_object(
        self,
        queryset: QuerySet[Collection] | None = None,  # noqa: ARG002, U100
    ) -> Collection:
        """Return the collection object to show."""
        return self.collection

    @override
    def get_context_data(self, *args: Any, **kwargs: Any) -> dict[str, Any]:
        """Return context_data with work_request_list and workspace_list."""
        context = super().get_context_data(**kwargs)
        if self.object.data:
            context["data"] = highlight_as_yaml(self.object.data)

        context["collections"] = (
            self.object.child_items.active()
            .filter(collection__isnull=False)
            .order_by("category", "name")
        )

        context["artifacts"] = (
            self.object.child_items.filter(artifact__isnull=False)
            .values("category")
            .annotate(
                count=Count("category", filter=Q(removed_at__isnull=True)),
                count_removed=Count(
                    "category", filter=Q(removed_at__isnull=False)
                ),
            )
            .order_by("category")
        )

        context["bare"] = (
            self.object.child_items.filter(
                collection__isnull=True, artifact__isnull=True
            )
            .values("category")
            .annotate(
                count=Count("category", filter=Q(removed_at__isnull=True)),
                count_removed=Count(
                    "category", filter=Q(removed_at__isnull=False)
                ),
            )
            .order_by("category")
        )

        for context_name, relation_type in (
            ("suite_based_on", CollectionRelation.Types.SUITE_BASED_ON),
            ("suite_targeting", CollectionRelation.Types.SUITE_TARGETING),
        ):
            try:
                context[context_name] = self.object.relations.get(
                    type=relation_type
                ).target
            except CollectionRelation.DoesNotExist:
                pass

        return context


class CollectionUpdateView(
    CollectionViewMixin, UpdateViewBase[Collection, CollectionUpdateForm]
):
    """Edit configuration for a collection."""

    form_class = CollectionUpdateForm
    template_name = "web/collection-update.html"

    @override
    def get_place(self) -> Place:
        return self.object.ui(self.request).place_update

    @override
    def get_object(
        self,
        queryset: QuerySet[Collection] | None = None,  # noqa: ARG002, U100
    ) -> Collection:
        """Return the collection object to update."""
        self.enforce(self.collection.can_configure)
        return self.collection


class CollectionDeleteView(
    CollectionViewMixin, DeleteViewBase[Collection, ModelFormBase[Collection]]
):
    """Delete a collection."""

    object: Collection
    model = Collection
    template_name = "web/collection-confirm_delete.html"

    @override
    def get_place(self) -> Place:
        return self.object.ui(self.request).place_delete

    @override
    def get_object(
        self,
        queryset: QuerySet[Collection] | None = None,  # noqa: ARG002, U100
    ) -> Collection:
        """Return the collection object to delete."""
        self.enforce(self.collection.can_delete)
        return self.collection

    @override
    def get_context_data(self, *args: Any, **kwargs: Any) -> dict[str, Any]:
        ctx = super().get_context_data(**kwargs)
        item_stats = list(self.object.get_item_stats())
        ctx["items_total"] = sum(x["count"] for x in item_stats)
        ctx["item_stats"] = item_stats
        return ctx

    @override
    def form_valid(
        self,
        form: ModelFormBase[Collection],  # noqa: U100
    ) -> HttpResponse:
        success_url = self.get_success_url()
        self.object.delete_collection()
        return HttpResponseRedirect(success_url)

    @override
    def get_success_url(self) -> str:
        return context.require_workspace().get_absolute_url()


class CollectionSearchView(
    TableMixin[CollectionItem],
    CollectionViewMixin,
    FormMixinBase[CollectionSearchForm],
    ListViewBase[CollectionItem],
):
    """Search a collection contents."""

    model = CollectionItem
    template_name = "web/collection-search.html"
    context_object_name = "item_list"
    form_class = CollectionSearchForm
    table_class = CollectionItemTable
    paginate_by = 50

    @override
    def get_place(self) -> Place:
        return self.collection.ui(self.request).place_search

    @override
    def get_queryset(self) -> QuerySet[CollectionItem]:
        """Filter items based on the search form."""
        queryset = (
            super().get_queryset().filter(parent_collection=self.collection)
        )
        queryset = queryset.select_related(
            "parent_collection__workspace__scope",
            "collection",
            "artifact__workspace__scope",
        )
        # FIXME: this builds the form twice: if it becomes a problem we can
        # override get_form to cache its result
        form = self.get_form()
        if form.is_valid():
            if name := form.cleaned_data["name"]:
                queryset = queryset.filter(name__contains=name)
            if category := form.cleaned_data["category"]:
                queryset = queryset.filter(category=category)
            if form.cleaned_data["historical"]:
                queryset = queryset.filter(removed_at__isnull=False)
            else:
                queryset = queryset.filter(removed_at__isnull=True)
        else:
            queryset = queryset.none()
        return queryset

    @override
    def get_form_kwargs(self) -> dict[str, Any]:
        kwargs = super().get_form_kwargs()
        kwargs["instance"] = self.collection
        kwargs["data"] = self.request.GET
        return kwargs

    @override
    def get_context_data(self, *args: Any, **kwargs: Any) -> dict[str, Any]:
        """Return context_data with work_request_list and workspace_list."""
        context = super().get_context_data(**kwargs)
        context["collection"] = self.collection
        return context


class CollectionItemDetailView(
    CollectionViewMixin, DetailViewBase[CollectionItem]
):
    """Show a collection item detail."""

    model = CollectionItem
    template_name = "web/collection-item-detail.html"
    context_object_name = "item"

    @override
    def get_place(self) -> Place:
        return self.object.ui(self.request).place

    @override
    def get_object(
        self, queryset: QuerySet[CollectionItem] | None = None
    ) -> CollectionItem:
        """Return the collection object to show."""
        assert queryset is None
        queryset = self.get_queryset()
        try:
            return queryset.filter(removed_at__isnull=True).get(
                pk=self.kwargs["iid"], parent_collection=self.collection
            )
        except CollectionItem.DoesNotExist:
            raise Http404(
                f"{self.kwargs['iid']} ({self.kwargs['iname']})"
                " item not found in "
                f"{self.kwargs['cname']}@{self.kwargs['ccat']}"
            )

    @override
    def get_context_data(self, *args: Any, **kwargs: Any) -> dict[str, Any]:
        context = super().get_context_data(**kwargs)
        if self.object.data:
            context["data"] = highlight_as_yaml(self.object.data)
        return context



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Common test-helper code involving wait tasks."""

from typing import Any, override

import debusine.tasks.tags as ttags
import debusine.worker.tags as wtags
from debusine.db.models import WorkRequest
from debusine.server.tasks.tests.helpers import DBTaskTestCase
from debusine.server.tasks.wait import BaseWaitTask
from debusine.server.workflows.models import WorkRequestWorkflowData
from debusine.test.django import TestCase


class WaitTaskTestCase[TaskClass: BaseWaitTask[Any, Any]](
    DBTaskTestCase[TaskClass], TestCase
):
    """Common base for wait task tests."""

    #: Set to true to mark the task as running on creation
    mark_running: bool

    @override
    def setUp(self) -> None:
        super().setUp()
        self.mark_running = False

    @override
    def work_request_create(self, **kwargs: Any) -> WorkRequest:
        task_data = self.task_data.copy()
        task_data.update(kwargs)
        work_request = self.playground.create_wait_task(
            task_name=self.task_class.name,
            task_data=self.task_class.task_data_type(**task_data),
            workflow_data=WorkRequestWorkflowData(needs_input=True),
        )
        if self.mark_running:
            self.playground.advance_work_request(
                work_request, mark_running=True
            )
        return work_request

    @override
    def build_expected_tag_candidates(
        self, **kwargs: set[str]
    ) -> tuple[ttags.TaskProvidedTags, wtags.TaskRequiredTags]:
        provided, required = super().build_expected_tag_candidates(**kwargs)
        required.add(
            wtags.ProvenanceRequired.SYSTEM, {wtags.WORKER_TYPE_NOT_ASSIGNABLE}
        )
        return provided, required
